#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from collections import defaultdict
from dyn.tm.errors import DynectGetError
from dyn.tm.services.dsf import DSFARecord, DSFAAAARecord, DSFCNAMERecord, \
    DSFFailoverChain, DSFMonitor, DSFNode, DSFRecordSet, DSFResponsePool, \
    DSFRuleset, TrafficDirector, get_all_dsf_monitors, get_all_dsf_services, \
    get_response_pool
from dyn.tm.session import DynectSession
from dyn.tm.zones import Zone as DynZone
from logging import getLogger
from threading import Lock
from uuid import uuid4

from ..record import Record, Update
from ..record.geo import GeoCodes
from .base import BaseProvider


###############################################################################
#
# The following monkey patching is to work around functionality that is lacking
# from DSFMonitor. You cannot set host or path (which we need) and there's no
# update method. What's more host & path aren't publically accessible on the
# object so you can't see their current values and depending on how the object
# came to be (constructor vs pulled from the api) the "private" location of
# those fields varies :-(
#
###############################################################################
def _monitor_host_get(self):
    return self._host or self._options['host']


DSFMonitor.host = property(_monitor_host_get)


def _monitor_host_set(self, value):
    if self._options is None:
        self._options = {}
    self._host = self._options['host'] = value


DSFMonitor.host = DSFMonitor.host.setter(_monitor_host_set)


def _monitor_path_get(self):
    return self._path or self._options['path']


DSFMonitor.path = property(_monitor_path_get)


def _monitor_path_set(self, value):
    if self._options is None:
        self._options = {}
    self._path = self._options['path'] = value


DSFMonitor.path = DSFMonitor.path.setter(_monitor_path_set)


def _monitor_protocol_get(self):
    return self._protocol


DSFMonitor.protocol = property(_monitor_protocol_get)


def _monitor_protocol_set(self, value):
    self._protocol = value


DSFMonitor.protocol = DSFMonitor.protocol.setter(_monitor_protocol_set)


def _monitor_port_get(self):
    return self._port or self._options['port']


DSFMonitor.port = property(_monitor_port_get)


def _monitor_port_set(self, value):
    if self._options is None:
        self._options = {}
    self._port = self._options['port'] = value


DSFMonitor.port = DSFMonitor.port.setter(_monitor_port_set)


def _monitor_update(self, host, path, protocol, port):
    # I can't see how to actually do this with the client lib so
    # I'm having to hack around it. Have to provide all the
    # options or else things complain
    return self._update({
        'protocol': protocol,
        'options': {
            'host': host,
            'path': path,
            'port': port,
            'timeout': DynProvider.MONITOR_TIMEOUT,
            'header': DynProvider.MONITOR_HEADER,
        }
    })


DSFMonitor.update = _monitor_update
###############################################################################


def _monitor_doesnt_match(monitor, host, path, protocol, port):
    return monitor.host != host or monitor.path != path or \
        monitor.protocol != protocol or int(monitor.port) != port


class _CachingDynZone(DynZone):
    log = getLogger('_CachingDynZone')

    _cache = {}

    @classmethod
    def get(cls, zone_name, create=False):
        cls.log.debug('get: zone_name=%s, create=%s', zone_name, create)
        # This works in dyn zone names, without the trailing .
        try:
            dyn_zone = cls._cache[zone_name]
            cls.log.debug('get:   cache hit')
        except KeyError:
            cls.log.debug('get:   cache miss')
            try:
                dyn_zone = _CachingDynZone(zone_name)
                cls.log.debug('get:   fetched')
            except DynectGetError:
                if not create:
                    cls.log.debug("get:   doesn't exist")
                    return None
                # this value shouldn't really matter, it's not tied to
                # whois or anything
                hostname = 'hostmaster@{}'.format(zone_name[:-1])
                # Try again with the params necessary to create
                dyn_zone = _CachingDynZone(zone_name, ttl=3600,
                                           contact=hostname,
                                           serial_style='increment')
                cls.log.debug('get:   created')
            cls._cache[zone_name] = dyn_zone

        return dyn_zone

    @classmethod
    def flush_zone(cls, zone_name):
        '''Flushes the zone cache, if there is one'''
        cls.log.debug('flush_zone: zone_name=%s', zone_name)
        try:
            del cls._cache[zone_name]
        except KeyError:
            pass

    def __init__(self, zone_name, *args, **kwargs):
        super(_CachingDynZone, self).__init__(zone_name, *args, **kwargs)
        self.flush_cache()

    def flush_cache(self):
        self._cached_records = None

    def get_all_records(self):
        if self._cached_records is None:
            self._cached_records = \
                super(_CachingDynZone, self).get_all_records()
        return self._cached_records

    def publish(self):
        super(_CachingDynZone, self).publish()
        self.flush_cache()


def _dynamic_value_sort_key(value):
    return value['value']


class DynProvider(BaseProvider):
    '''
    Dynect Managed DNS provider

    dyn:
        class: octodns.provider.dyn.DynProvider
        # Your dynect customer name (required)
        customer: cust
        # Your dynect username (required)
        username: user
        # Your dynect password (required)
        password: pass
        # Whether or not to support TrafficDirectors and enable GeoDNS
        # (optional, default is false)
        traffic_directors_enabled: true

    Note: due to the way dyn.tm.session.DynectSession is managing things we can
    only really have a single DynProvider configured. When you create a
    DynectSession it's stored in a thread-local singleton. You don't invoke
    methods on this session or a client that holds on to it. The client
    libraries grab their per-thread session by accessing the singleton through
    DynectSession.get_session(). That fundamentally doesn't support having more
    than one account active at a time. See DynProvider._check_dyn_sess for some
    related bits.
    '''

    RECORDS_TO_TYPE = {
        'a_records': 'A',
        'aaaa_records': 'AAAA',
        'alias_records': 'ALIAS',
        'caa_records': 'CAA',
        'cname_records': 'CNAME',
        'mx_records': 'MX',
        'naptr_records': 'NAPTR',
        'ns_records': 'NS',
        'ptr_records': 'PTR',
        'sshfp_records': 'SSHFP',
        'spf_records': 'SPF',
        'srv_records': 'SRV',
        'txt_records': 'TXT',
    }
    TYPE_TO_RECORDS = {v: k for k, v in RECORDS_TO_TYPE.items()}
    SUPPORTS = set(TYPE_TO_RECORDS.keys())

    # https://help.dyn.com/predefined-geotm-regions-groups/
    REGION_CODES = {
        'NA': 11,  # Continental North America
        'SA': 12,  # Continental South America
        'EU': 13,  # Continental Europe
        'AF': 14,  # Continental Africa
        'AS': 15,  # Continental Asia
        'OC': 16,  # Continental Australia/Oceania
        'AN': 17,  # Continental Antarctica
    }
    # Reverse of ^
    REGION_CODES_LOOKUP = {code: geo for geo, code in REGION_CODES.items()}

    MONITOR_HEADER = 'User-Agent: Dyn Monitor'
    MONITOR_TIMEOUT = 10

    _sess_create_lock = Lock()

    def __init__(self, id, customer, username, password,
                 traffic_directors_enabled=False, *args, **kwargs):
        self.log = getLogger('DynProvider[{}]'.format(id))
        self.log.debug('__init__: id=%s, customer=%s, username=%s, '
                       'password=***, traffic_directors_enabled=%s', id,
                       customer, username, traffic_directors_enabled)
        # we have to set this before calling super b/c SUPPORTS_GEO requires it
        self.traffic_directors_enabled = traffic_directors_enabled
        super(DynProvider, self).__init__(id, *args, **kwargs)
        self.customer = customer
        self.username = username
        self.password = password

        self._cache = {}
        self._traffic_directors = None
        self._traffic_director_monitors = None

    @property
    def SUPPORTS_GEO(self):
        return self.traffic_directors_enabled

    @property
    def SUPPORTS_DYNAMIC(self):
        return self.traffic_directors_enabled

    def _check_dyn_sess(self):
        # We don't have to worry about locking for the check since the
        # underlying pieces are pre-thread. We can check to see if this thread
        # has a session and if so we're good to go.
        if DynectSession.get_session() is None:
            # We need to create a new session for this thread and DynectSession
            # creation is not thread-safe so we have to do the locking. If we
            # don't and multiple sessions start creation before the the first
            # has finished (long time b/c it makes http calls) the subsequent
            # creates will blow away DynectSession._instances, potentially
            # multiple times if there are multiple creates in flight. Only the
            # last of these initial concurrent creates will exist in
            # DynectSession._instances dict and the others will be lost. When
            # this thread later tries to make api calls there won't be an
            # accessible session available for it to use.
            with self._sess_create_lock:
                DynectSession(self.customer, self.username, self.password)

    def _data_for_A(self, _type, records):
        return {
            'type': _type,
            'ttl': records[0].ttl,
            'values': [r.address for r in records]
        }

    _data_for_AAAA = _data_for_A

    def _data_for_ALIAS(self, _type, records):
        # See note on ttl in _kwargs_for_ALIAS
        record = records[0]
        return {
            'type': _type,
            'ttl': record.ttl,
            'value': record.alias
        }

    def _data_for_CAA(self, _type, records):
        return {
            'type': _type,
            'ttl': records[0].ttl,
            'values': [{'flags': r.flags, 'tag': r.tag, 'value': r.value}
                       for r in records],
        }

    def _data_for_CNAME(self, _type, records):
        record = records[0]
        return {
            'type': _type,
            'ttl': record.ttl,
            'value': record.cname,
        }

    def _data_for_MX(self, _type, records):
        return {
            'type': _type,
            'ttl': records[0].ttl,
            'values': [{'preference': r.preference, 'exchange': r.exchange}
                       for r in records],
        }

    def _data_for_NAPTR(self, _type, records):
        return {
            'type': _type,
            'ttl': records[0].ttl,
            'values': [{
                'order': r.order,
                'preference': r.preference,
                'flags': r.flags,
                'service': r.services,
                'regexp': r.regexp,
                'replacement': r.replacement,
            } for r in records]
        }

    def _data_for_NS(self, _type, records):
        return {
            'type': _type,
            'ttl': records[0].ttl,
            'values': [r.nsdname for r in records]
        }

    def _data_for_PTR(self, _type, records):
        record = records[0]
        return {
            'type': _type,
            'ttl': record.ttl,
            'value': record.ptrdname,
        }

    def _data_for_SPF(self, _type, records):
        record = records[0]
        return {
            'type': _type,
            'ttl': record.ttl,
            'values': [r.txtdata for r in records]
        }

    _data_for_TXT = _data_for_SPF

    def _data_for_SSHFP(self, _type, records):
        return {
            'type': _type,
            'ttl': records[0].ttl,
            'values': [{
                'algorithm': r.algorithm,
                'fingerprint_type': r.fptype,
                'fingerprint': r.fingerprint,
            } for r in records],
        }

    def _data_for_SRV(self, _type, records):
        return {
            'type': _type,
            'ttl': records[0].ttl,
            'values': [{
                'priority': r.priority,
                'weight': r.weight,
                'port': r.port,
                'target': r.target,
            } for r in records],
        }

    @property
    def traffic_directors(self):
        if self._traffic_directors is None:
            self._check_dyn_sess()

            tds = defaultdict(dict)
            for td in get_all_dsf_services():
                try:
                    fqdn, _type = td.label.split(':', 1)
                except ValueError:
                    self.log.warn("Unsupported TrafficDirector '%s'", td.label)
                    continue
                tds[fqdn][_type] = td
            self._traffic_directors = dict(tds)

        return self._traffic_directors

    def _populate_geo_traffic_director(self, zone, fqdn, _type, td, rulesets,
                                       lenient):
        # We start out with something that will always show change in case this
        # is a busted TD. This will prevent us from creating a duplicate td.
        # We'll overwrite this with real data provided we have it
        geo = {}
        data = {
            'geo': geo,
            'type': _type,
            'ttl': td.ttl,
            'values': ['0.0.0.0']
        }
        for ruleset in rulesets:
            try:
                record_set = ruleset.response_pools[0].rs_chains[0] \
                    .record_sets[0]
            except IndexError:
                # problems indicate a malformed ruleset, ignore it
                continue
            if ruleset.label.startswith('default:'):
                data_for = getattr(self, '_data_for_{}'.format(_type))
                data.update(data_for(_type, record_set.records))
            else:
                # We've stored the geo in label
                try:
                    code, _ = ruleset.label.split(':', 1)
                except ValueError:
                    continue
                values = [r.address for r in record_set.records]
                geo[code] = values

        name = zone.hostname_from_fqdn(fqdn)
        record = Record.new(zone, name, data, source=self)
        zone.add_record(record, lenient=lenient)

        return record

    def _value_for_address(self, _type, record):
        return {
            'value': record.address,
            'weight': record.weight,
        }

    _value_for_A = _value_for_address
    _value_for_AAAA = _value_for_address

    def _value_for_CNAME(self, _type, record):
        return {
            'value': record.cname,
            'weight': record.weight,
        }

    def _populate_dynamic_pools(self, _type, rulesets, response_pools):
        default = {}
        pools = {}

        data_for = getattr(self, '_data_for_{}'.format(_type))
        value_for = getattr(self, '_value_for_{}'.format(_type))

        # Build the list of pools, we can't just read them off of rules b/c we
        # won't see unused pools there. If/when we dis-allow unused pools we
        # could probably change that and avoid the refresh
        for response_pool in response_pools:
            # We have to refresh the response pool to have access to its
            # rs_chains and thus records, yeah... :-(
            # TODO: look at rulesets first b/c they won't need a refresh...
            response_pool.refresh()
            try:
                record_set = response_pool.rs_chains[0] \
                    .record_sets[0]
            except IndexError:
                # problems indicate a malformed ruleset, ignore it
                self.log.warn('_populate_dynamic_pools: '
                              'malformed response_pool "%s" ignoring',
                              response_pool.label)
                continue

            label = response_pool.label

            if label == 'default':
                # The default pool has the base record values
                default = data_for(_type, record_set.records)
            else:
                if label not in pools:
                    # First time we've seen it get its data
                    # Note we'll have to set fallbacks as we go through rules
                    # b/c we can't determine them here
                    values = [value_for(_type, r) for r in record_set.records]
                    # Sort to ensure consistent ordering so we can compare them
                    values.sort(key=_dynamic_value_sort_key)
                    pools[label] = {
                        'values': values,
                    }

        return default, pools

    def _populate_dynamic_rules(self, rulesets, pools):
        rules = []

        # Build the list of rules based on the rulesets
        for ruleset in rulesets:
            if ruleset.label.startswith('default:'):
                # Ignore the default, it's implicit in our model
                continue

            num_pools = len(ruleset.response_pools)
            if num_pools > 0:
                # Find the primary pool for this rule
                pool = ruleset.response_pools[0].label
                # TODO: verify pool exists
                if num_pools > 1:
                    # We have a fallback, record it in the approrpriate pool.
                    # Note we didn't have fallback info when we populated the
                    # pools above so we're filling that info in here. It's
                    # possible that rules will have disagreeing values for the
                    # fallbacks. That's annoying but a sync should fix it and
                    # match stuff up with the config.
                    fallback = ruleset.response_pools[1].label
                    # TODO: verify fallback exists
                    if fallback != 'default':
                        pools[pool]['fallback'] = fallback
            else:
                self.log.warn('_populate_dynamic_pools: '
                              'ruleset "%s" has no response_pools',
                              ruleset.label)
                continue

            # OK we have the rule's pool info, record it and work on the rule's
            # matching criteria
            rule = {
                'pool': pool,
            }

            criteria_type = ruleset.criteria_type
            if criteria_type == 'geoip':
                # Geo
                geo = ruleset.criteria['geoip']
                geos = []
                # Dyn uses the same 2-letter codes as octoDNS (except for
                # continents) but it doesn't have the hierary, e.g. US is
                # just US, not NA-US. We'll have to map these things back
                for code in geo['country']:
                    geos.append(GeoCodes.country_to_code(code))
                for code in geo['province']:
                    geos.append(GeoCodes.province_to_code(code.upper()))
                for code in geo['region']:
                    geos.append(self.REGION_CODES_LOOKUP[int(code)])
                geos.sort()
                rule['geos'] = geos
            elif criteria_type == 'always':
                pass
            else:
                self.log.warn('_populate_dynamic_rules: '
                              'unsupported criteria_type "%s", ignoring',
                              criteria_type)
                continue

            rules.append(rule)

        return rules

    def _populate_dynamic_traffic_director(self, zone, fqdn, _type, td,
                                           rulesets, lenient):
        # We'll go ahead and grab pools too, using all will include unref'd
        # pools
        response_pools = td.all_response_pools

        # Populate pools
        default, pools = self._populate_dynamic_pools(_type, rulesets,
                                                      response_pools)

        # Populate rules
        rules = self._populate_dynamic_rules(rulesets, pools)

        # We start out with something that will always show
        # change in case this is a busted TD. This will prevent us from
        # creating a duplicate td. We'll overwrite this with real data
        # provide we have it
        data = {
            'dynamic': {
                'pools': pools,
                'rules': rules,
            },
            'type': _type,
            'ttl': td.ttl,
        }
        # Include default's information in data
        data.update(default)

        name = zone.hostname_from_fqdn(fqdn)
        record = Record.new(zone, name, data, source=self, lenient=lenient)
        zone.add_record(record, lenient=lenient)

        return record

    def _is_traffic_director_dyanmic(self, td, rulesets):
        for ruleset in rulesets:
            try:
                pieces = ruleset.label.split(':')
                if len(pieces) == 2:
                    # It matches octoDNS's format
                    int(pieces[0])
                    # It's an integer, so probably rule_num, thus dynamic
                    return True
            except (IndexError, ValueError):
                pass
        # We didn't see any rulesets that look like a dynamic record so maybe
        # geo...
        return False

    def _populate_traffic_directors(self, zone, lenient):
        self.log.debug('_populate_traffic_directors: zone=%s, lenient=%s',
                       zone.name, lenient)
        td_records = set()
        for fqdn, types in self.traffic_directors.items():
            for _type, td in types.items():
                # Does this TD belong to the current zone
                td_zone = '{}.'.format(td.nodes[0]['zone'])
                if td_zone != zone.name:
                    # Doesn't belong to the current zone, skip it
                    continue
                # critical to call rulesets once, each call loads them :-(
                rulesets = td.rulesets
                if self._is_traffic_director_dyanmic(td, rulesets):
                    record = \
                        self._populate_dynamic_traffic_director(zone, fqdn,
                                                                _type, td,
                                                                rulesets,
                                                                lenient)
                else:
                    record = \
                        self._populate_geo_traffic_director(zone, fqdn, _type,
                                                            td, rulesets,
                                                            lenient)
                td_records.add(record)

        return td_records

    def populate(self, zone, target=False, lenient=False):
        self.log.debug('populate: name=%s, target=%s, lenient=%s', zone.name,
                       target, lenient)

        exists = False
        before = len(zone.records)

        self._check_dyn_sess()

        td_records = set()
        if self.traffic_directors_enabled:
            td_records = self._populate_traffic_directors(zone, lenient)
            exists = True

        dyn_zone = _CachingDynZone.get(zone.name[:-1])

        if dyn_zone:
            exists = True
            values = defaultdict(lambda: defaultdict(list))
            for _type, records in dyn_zone.get_all_records().items():
                if _type == 'soa_records':
                    continue
                _type = self.RECORDS_TO_TYPE[_type]
                for record in records:
                    record_name = zone.hostname_from_fqdn(record.fqdn)
                    values[record_name][_type].append(record)

            for name, types in values.items():
                for _type, records in types.items():
                    data_for = getattr(self, '_data_for_{}'.format(_type))
                    data = data_for(_type, records)
                    record = Record.new(zone, name, data, source=self,
                                        lenient=lenient)
                    if record not in td_records:
                        zone.add_record(record, lenient=lenient)

        self.log.info('populate:   found %s records, exists=%s',
                      len(zone.records) - before, exists)
        return exists

    def _extra_changes(self, desired, changes, **kwargs):
        self.log.debug('_extra_changes: desired=%s', desired.name)

        changed = set([c.record for c in changes])

        extra = []
        for record in desired.records:
            if record in changed or not getattr(record, 'geo', False):
                # Already changed, or no geo, no need to check it
                continue
            label = '{}:{}'.format(record.fqdn, record._type)
            try:
                monitor = self.traffic_director_monitors[label]
            except KeyError:
                self.log.info('_extra_changes: health-check missing for %s',
                              label)
                extra.append(Update(record, record))
                continue
            if _monitor_doesnt_match(monitor, record.healthcheck_host,
                                     record.healthcheck_path,
                                     record.healthcheck_protocol,
                                     record.healthcheck_port):
                self.log.info('_extra_changes: health-check mis-match for %s',
                              label)
                extra.append(Update(record, record))

        return extra

    def _kwargs_for_A(self, record):
        return [{
            'address': v,
            'ttl': record.ttl,
        } for v in record.values]

    _kwargs_for_AAAA = _kwargs_for_A

    def _kwargs_for_CAA(self, record):
        return [{
            'flags': v.flags,
            'tag': v.tag,
            'ttl': record.ttl,
            'value': v.value,
        } for v in record.values]

    def _kwargs_for_CNAME(self, record):
        return [{
            'cname': record.value,
            'ttl': record.ttl,
        }]

    def _kwargs_for_ALIAS(self, record):
        # NOTE: Dyn's UI doesn't allow editing of ALIAS ttl, but the API seems
        # to accept and store the values we send it just fine. No clue if they
        # do anything with them. I'd assume they just obey the TTL of the
        # record that we're pointed at which makes sense.
        return [{
            'alias': record.value,
            'ttl': record.ttl,
        }]

    def _kwargs_for_MX(self, record):
        return [{
            'preference': v.preference,
            'exchange': v.exchange,
            'ttl': record.ttl,
        } for v in record.values]

    def _kwargs_for_NAPTR(self, record):
        return [{
            'flags': v.flags,
            'order': v.order,
            'preference': v.preference,
            'regexp': v.regexp,
            'replacement': v.replacement,
            'services': v.service,
            'ttl': record.ttl,
        } for v in record.values]

    def _kwargs_for_NS(self, record):
        return [{
            'nsdname': v,
            'ttl': record.ttl,
        } for v in record.values]

    def _kwargs_for_PTR(self, record):
        return [{
            'ptrdname': record.value,
            'ttl': record.ttl,
        }]

    def _kwargs_for_SSHFP(self, record):
        return [{
            'algorithm': v.algorithm,
            'fptype': v.fingerprint_type,
            'fingerprint': v.fingerprint,
        } for v in record.values]

    def _kwargs_for_SPF(self, record):
        return [{
            'txtdata': v,
            'ttl': record.ttl,
        } for v in record.chunked_values]

    def _kwargs_for_SRV(self, record):
        return [{
            'port': v.port,
            'priority': v.priority,
            'target': v.target,
            'weight': v.weight,
            'ttl': record.ttl,
        } for v in record.values]

    _kwargs_for_TXT = _kwargs_for_SPF

    @property
    def traffic_director_monitors(self):
        if self._traffic_director_monitors is None:
            self.log.debug('traffic_director_monitors: loading')
            self._traffic_director_monitors = \
                {m.label: m for m in get_all_dsf_monitors()}

        return self._traffic_director_monitors

    def _traffic_director_monitor(self, record):
        fqdn = record.fqdn
        label = '{}:{}'.format(fqdn, record._type)
        try:
            try:
                monitor = self.traffic_director_monitors[label]
                self.log.debug('_traffic_director_monitor: existing for %s',
                               label)
            except KeyError:
                # UNTIL 1.0 We don't have one for the new label format, see if
                # we still have one for the old and update it
                monitor = self.traffic_director_monitors[fqdn]
                self.log.info('_traffic_director_monitor: upgrading label '
                              'to %s', label)
                monitor.label = label
                self.traffic_director_monitors[label] = \
                    self.traffic_director_monitors[fqdn]
                del self.traffic_director_monitors[fqdn]
            if _monitor_doesnt_match(monitor, record.healthcheck_host,
                                     record.healthcheck_path,
                                     record.healthcheck_protocol,
                                     record.healthcheck_port):
                self.log.info('_traffic_director_monitor: updating monitor '
                              'for %s', label)
                monitor.update(record.healthcheck_host,
                               record.healthcheck_path,
                               record.healthcheck_protocol,
                               record.healthcheck_port)
            return monitor
        except KeyError:
            self.log.info('_traffic_director_monitor: creating monitor '
                          'for %s', label)
            monitor = DSFMonitor(label, protocol=record.healthcheck_protocol,
                                 response_count=2, probe_interval=60,
                                 retries=2, port=record.healthcheck_port,
                                 active='Y', host=record.healthcheck_host,
                                 timeout=self.MONITOR_TIMEOUT,
                                 header=self.MONITOR_HEADER,
                                 path=record.healthcheck_path)
            self._traffic_director_monitors[label] = monitor
            return monitor

    def _find_or_create_geo_pool(self, td, pools, label, _type, values,
                                 monitor_id=None):
        for pool in pools:
            if pool.label != label:
                continue
            records = pool.rs_chains[0].record_sets[0].records
            record_values = sorted([r.address for r in records])
            if record_values == values:
                # it's a match
                return pool
        # we need to create the pool
        _class = {
            'A': DSFARecord,
            'AAAA': DSFAAAARecord
        }[_type]
        records = [_class(v) for v in values]
        record_set = DSFRecordSet(_type, label, serve_count=len(records),
                                  records=records, dsf_monitor_id=monitor_id)
        chain = DSFFailoverChain(label, record_sets=[record_set])
        pool = DSFResponsePool(label, rs_chains=[chain])
        pool.create(td)

        # We need to store the newly created pool in the pools list since the
        # caller won't know if it was newly created or not. This will allow us
        # to find this pool again if another rule references it and avoid
        # creating duplicates
        pools.append(pool)

        return pool

    def _dynamic_records_for_A(self, values, record_extras):
        return [DSFARecord(v['value'], weight=v.get('weight', 1),
                           **record_extras)
                for v in values]

    def _dynamic_records_for_AAAA(self, values, record_extras):
        return [DSFAAAARecord(v['value'], weight=v.get('weight', 1),
                              **record_extras)
                for v in values]

    def _dynamic_records_for_CNAME(self, values, record_extras):
        return [DSFCNAMERecord(v['value'], weight=v.get('weight', 1),
                               **record_extras)
                for v in values]

    def _find_or_create_dynamic_pool(self, td, pools, label, _type, values,
                                     monitor_id=None, record_extras={}):

        # Sort the values for consistent ordering so that we can compare
        values = sorted(values, key=_dynamic_value_sort_key)
        # Ensure that weight is included and if not use the default
        values = map(lambda v: {
            'value': v['value'],
            'weight': v.get('weight', 1),
        }, values)

        # Walk through our existing pools looking for a match we can use
        for pool in pools:
            # It must have the same label
            if pool.label != label:
                continue
            try:
                records = pool.rs_chains[0].record_sets[0].records
            except IndexError:
                # No values, can't match
                continue
            # And the (sorted) values must match once converted for comparison
            # purposes
            value_for = getattr(self, '_value_for_{}'.format(_type))
            record_values = [value_for(_type, r) for r in records]
            if record_values == values:
                # it's a match
                return pool

        # We don't have this pool and thus need to create it
        records_for = getattr(self, '_dynamic_records_for_{}'.format(_type))
        records = records_for(values, record_extras)
        record_set = DSFRecordSet(_type, label, serve_count=1, records=records,
                                  dsf_monitor_id=monitor_id)
        chain = DSFFailoverChain(label, record_sets=[record_set])
        pool = DSFResponsePool(label, rs_chains=[chain])
        pool.create(td)

        # We need to store the newly created pool in the pools list since the
        # caller won't know if it was newly created or not. This will allow us
        # to find this pool again if another rule references it and avoid
        # creating duplicates
        pools.append(pool)

        return pool

    def _mod_geo_rulesets(self, td, change):
        new = change.new

        # Response Pools
        pools = {}

        # Get existing pools. This should be simple, but it's not b/c the dyn
        # api is a POS. We need all response pools so we can GC and check to
        # make sure that what we're after doesn't already exist.
        # td.all_response_pools just returns thin objects that don't include
        # their rs_chains (and children down to actual records.) We could just
        # foreach over those turning them into full DSFResponsePool objects
        # with get_response_pool, but that'd be N round-trips. We can avoid
        # those round trips in cases where the pools are in use in rules where
        # they're already full objects.

        # First up populate all the full pools we have under rules, the _
        # prevents a td.refresh we don't need :-( seriously?
        existing_rulesets = td._rulesets
        for ruleset in existing_rulesets:
            for pool in ruleset.response_pools:
                pools[pool.response_pool_id] = pool
        # Reverse sort the existing_rulesets by _ordering so that we'll remove
        # them in that order later, this will ensure that we remove the old
        # default before any of the old geo rules preventing it from catching
        # everything.
        existing_rulesets.sort(key=lambda r: r._ordering, reverse=True)

        # Now we need to find any pools that aren't referenced by rules
        for pool in td.all_response_pools:
            rpid = pool.response_pool_id
            if rpid not in pools:
                # we want this one, but it's thin, inflate it
                pools[rpid] = get_response_pool(rpid, td)
        # now that we have full objects for the complete set of existing pools,
        # a list will be more useful
        pools = pools.values()

        # Rulesets

        # We need to make sure and insert the new rules after any existing
        # rules so they won't take effect before we've had a chance to add
        # response pools to them. I've tried both publish=False (which is
        # completely broken in the client) and creating the rulesets with
        # response_pool_ids neither of which appear to work from the client
        # library. If there are no existing rulesets fallback to 0
        insert_at = max([
            int(r._ordering)
            for r in existing_rulesets
        ] + [-1]) + 1
        self.log.debug('_mod_geo_rulesets: insert_at=%d', insert_at)

        # add the default
        label = 'default:{}'.format(uuid4().hex)
        ruleset = DSFRuleset(label, 'always', [])
        ruleset.create(td, index=insert_at)
        pool = self._find_or_create_geo_pool(td, pools, 'default', new._type,
                                             new.values)
        # There's no way in the client lib to create a ruleset with an existing
        # pool (ref'd by id) so we have to do this round-a-bout.
        active_pools = {
            'default': pool.response_pool_id
        }
        ruleset.add_response_pool(pool.response_pool_id)

        monitor_id = self._traffic_director_monitor(new).dsf_monitor_id
        # Geos ordered least to most specific so that parents will always be
        # created before their children (and thus can be referenced
        geos = sorted(new.geo.items(), key=lambda d: d[0])
        for _, geo in geos:
            if geo.subdivision_code:
                criteria = {
                    'province': geo.subdivision_code.lower()
                }
            elif geo.country_code:
                criteria = {
                    'country': geo.country_code
                }
            else:
                criteria = {
                    'region': self.REGION_CODES[geo.continent_code]
                }

            label = '{}:{}'.format(geo.code, uuid4().hex)
            ruleset = DSFRuleset(label, 'geoip', [], {
                'geoip': criteria
            })
            # Something you have to call create others the constructor does it
            ruleset.create(td, index=insert_at)

            first = geo.values[0]
            pool = self._find_or_create_geo_pool(td, pools, first, new._type,
                                                 geo.values, monitor_id)
            active_pools[geo.code] = pool.response_pool_id
            ruleset.add_response_pool(pool.response_pool_id)

            # look for parent rulesets we can add in the chain
            for code in geo.parents:
                try:
                    pool_id = active_pools[code]
                    # looking at client lib code, index > exists appends
                    ruleset.add_response_pool(pool_id, index=999)
                except KeyError:
                    pass
            # and always add default as the last
            pool_id = active_pools['default']
            ruleset.add_response_pool(pool_id, index=999)

        # we're done with active_pools as a lookup, convert it in to a set of
        # the ids in use
        active_pools = set(active_pools.values())
        # Clean up unused response_pools
        for pool in pools:
            if pool.response_pool_id in active_pools:
                continue
            pool.delete()

        # Clean out the old rulesets
        for ruleset in existing_rulesets:
            ruleset.delete()

    def _mod_geo_Create(self, dyn_zone, change):
        new = change.new
        fqdn = new.fqdn
        _type = new._type
        label = '{}:{}'.format(fqdn, _type)
        node = DSFNode(new.zone.name, fqdn)
        td = TrafficDirector(label, ttl=new.ttl, nodes=[node], publish='Y')
        self.log.debug('_mod_geo_Create: td=%s', td.service_id)
        self._mod_geo_rulesets(td, change)
        self.traffic_directors[fqdn] = {
            _type: td
        }

    def _mod_geo_Update(self, dyn_zone, change):
        new = change.new
        if not new.geo:
            # New record doesn't have geo we're going from a TD to a regular
            # record
            self._mod_Create(dyn_zone, change)
            self._mod_geo_Delete(dyn_zone, change)
            return
        try:
            td = self.traffic_directors[new.fqdn][new._type]
        except KeyError:
            # There's no td, this is actually a create, we must be going from a
            # non-geo to geo record so delete the regular record as well
            self._mod_geo_Create(dyn_zone, change)
            self._mod_Delete(dyn_zone, change)
            return
        self._mod_geo_rulesets(td, change)

    def _mod_geo_Delete(self, dyn_zone, change):
        existing = change.existing
        fqdn_tds = self.traffic_directors[existing.fqdn]
        _type = existing._type
        fqdn_tds[_type].delete()
        del fqdn_tds[_type]

    def _mod_dynamic_rulesets(self, td, change):
        new = change.new

        # TODO: make sure we can update TTLs
        if td.ttl != new.ttl:
            td.ttl = new.ttl

        # Get existing pools. This should be simple, but it's not b/c the dyn
        # api is a POS. We need all response pools so we can GC and check to
        # make sure that what we're after doesn't already exist.
        # td.all_response_pools just returns thin objects that don't include
        # their rs_chains (and children down to actual records.) We could just
        # foreach over those turning them into full DSFResponsePool objects
        # with get_response_pool, but that'd be N round-trips. We can avoid
        # those round trips in cases where the pools are in use in rules where
        # they're already full objects.

        # First up populate all the pools we have under rules, the _ prevents a
        # td.refresh we don't need :-( seriously?
        existing_rulesets = td._rulesets
        pools = {}
        for ruleset in existing_rulesets:
            for pool in ruleset.response_pools:
                pools[pool.response_pool_id] = pool

        # Reverse sort the existing_rulesets by _ordering so that we'll remove
        # them in that order later, this will ensure that we remove the old
        # default before any of the old geo rules preventing it from catching
        # everything.
        existing_rulesets.sort(key=lambda r: r._ordering, reverse=True)

        # Add in any pools that aren't currently referenced by rules
        for pool in td.all_response_pools:
            rpid = pool.response_pool_id
            if rpid not in pools:
                # we want this one, but it's thin, inflate it
                pools[rpid] = get_response_pool(rpid, td)
        # now that we have full objects for the complete set of existing pools,
        # a list will be more useful
        pools = pools.values()

        # Rulesets

        # We need to make sure and insert the new rules after any existing
        # rules so they won't take effect before we've had a chance to add
        # response pools to them. I've tried both publish=False (which is
        # completely broken in the client) and creating the rulesets with
        # response_pool_ids neither of which appear to work from the client
        # library. If there are no existing rulesets fallback to 0
        insert_at = max([
            int(r._ordering)
            for r in existing_rulesets
        ] + [-1]) + 1
        self.log.debug('_mod_dynamic_rulesets: insert_at=%d', insert_at)

        # Add the base record values as the ultimate/unhealthchecked default
        label = 'default:{}'.format(uuid4().hex)
        ruleset = DSFRuleset(label, 'always', [])
        ruleset.create(td, index=insert_at)
        # If/when we go beyond A, AAAA, and CNAME this will have to get
        # more intelligent, probably a weighted_values method on Record objects
        # or something like that?
        try:
            values = new.values
        except AttributeError:
            values = [new.value]
        values = [{
            'value': v,
            'weight': 1,
        } for v in values]
        # For these defaults we need to set them to always be served and to
        # ignore any health checking (since they won't have one)
        pool = self._find_or_create_dynamic_pool(td, pools, 'default',
                                                 new._type, values,
                                                 record_extras={
                                                     'automation': 'manual',
                                                     'eligible': True,
                                                 })
        # There's no way in the client lib to create a ruleset with an existing
        # pool (ref'd by id) so we have to do this round-a-bout.
        active_pools = {
            # TODO: disallow default as a pool id
            'default': pool.response_pool_id
        }
        ruleset.add_response_pool(pool.response_pool_id)

        # Get our monitor
        monitor_id = self._traffic_director_monitor(new).dsf_monitor_id

        # Make sure we have all the pools we're going to need
        for _id, pool in sorted(new.dynamic.pools.items()):
            values = [{
                'weight': v.get('weight', 1),
                'value': v['value'],
            } for v in pool.data['values']]
            pool = self._find_or_create_dynamic_pool(td, pools, _id,
                                                     new._type, values,
                                                     monitor_id)
            active_pools[_id] = pool.response_pool_id

        # Run through and configure our rules
        for rule_num, rule in enumerate(reversed(new.dynamic.rules)):
            criteria = defaultdict(lambda: defaultdict(list))
            criteria_type = 'always'
            try:
                geos = rule.data['geos']
                criteria_type = 'geoip'
            except KeyError:
                geos = []

            for geo in geos:
                geo = GeoCodes.parse(geo)
                if geo['province_code']:
                    criteria['geoip']['province'] \
                        .append(geo['province_code'].lower())
                elif geo['country_code']:
                    criteria['geoip']['country'] \
                        .append(geo['country_code'])
                else:
                    criteria['geoip']['region'] \
                        .append(self.REGION_CODES[geo['continent_code']])

            label = '{}:{}'.format(rule_num, uuid4().hex)
            ruleset = DSFRuleset(label, criteria_type, [], criteria)
            # Something you have to call create others the constructor does it
            ruleset.create(td, index=insert_at)

            # Add the primary pool for this rule
            rule_pool = rule.data['pool']
            ruleset.add_response_pool(active_pools[rule_pool])

            # OK, we have the rule and its primary pool setup, now look to see
            # if there's a fallback chain that needs to be configured
            fallback = new.dynamic.pools[rule_pool].data.get('fallback', None)
            seen = set([rule_pool])
            while fallback and fallback not in seen:
                seen.add(fallback)
                # looking at client lib code, index > exists appends
                ruleset.add_response_pool(active_pools[fallback], index=999)
                fallback = new.dynamic.pools[fallback].data.get('fallback',
                                                                None)
            if fallback is not None:
                # If we're out of the while and fallback is not None that means
                # there was a loop. This generally shouldn't happen since
                # Record validations test for it, but this is a
                # belt-and-suspenders setup. Excepting here would put things
                # into a partially configured state which would be bad. We'll
                # just break at the point where the loop was going to happen
                # and log about it. Note that any time we hit this we're likely
                # to hit it multiple times as we configure the other pools
                self.log.warn('_mod_dynamic_rulesets: loop detected in '
                              'fallback chain, fallback=%s, seen=%s', fallback,
                              seen)

            # and always add default as the last
            ruleset.add_response_pool(active_pools['default'], index=999)

        # we're done with active_pools as a lookup, convert it in to a set of
        # the ids in use
        active_pools = set(active_pools.values())
        # Clean up unused response_pools
        for pool in pools:
            if pool.response_pool_id in active_pools:
                continue
            pool.delete()

        # Clean out the old rulesets
        for ruleset in existing_rulesets:
            ruleset.delete()

    def _mod_dynamic_Create(self, dyn_zone, change):
        new = change.new
        fqdn = new.fqdn
        _type = new._type
        # Create a new traffic director
        label = '{}:{}'.format(fqdn, _type)
        node = DSFNode(new.zone.name, fqdn)
        td = TrafficDirector(label, ttl=new.ttl, nodes=[node], publish='Y')
        self.log.debug('_mod_dynamic_Create: td=%s', td.service_id)
        # Sync up it's pools & rules
        self._mod_dynamic_rulesets(td, change)
        # Store it for future reference
        self.traffic_directors[fqdn] = {
            _type: td
        }

    def _mod_dynamic_Update(self, dyn_zone, change):
        new = change.new
        if not new.dynamic:
            if new.geo:
                # New record is a geo record
                self.log.info('_mod_dynamic_Update: %s to geo', new.fqdn)
                # Convert the TD over to a geo and we're done
                self._mod_geo_Update(dyn_zone, change)
            else:
                # New record doesn't have dynamic, we're going from a TD to a
                # regular record
                self.log.info('_mod_dynamic_Update: %s to plain', new.fqdn)
                # Create the regular record
                self._mod_Create(dyn_zone, change)
                # Delete the dynamic
                self._mod_dynamic_Delete(dyn_zone, change)
            return
        try:
            # We'll be dynamic going forward, see if we have one already
            td = self.traffic_directors[new.fqdn][new._type]
            if change.existing.geo:
                self.log.info('_mod_dynamic_Update: %s from geo', new.fqdn)
            else:
                self.log.debug('_mod_dynamic_Update: %s existing', new.fqdn)
            # If we're here we do, we'll just update it down below
        except KeyError:
            # There's no td, this is actually a create, we must be going from a
            # non-dynamic to dynamic record
            # First create the dynamic record
            self.log.info('_mod_dynamic_Update: %s from regular', new.fqdn)
            self._mod_dynamic_Create(dyn_zone, change)
            # From a generic so remove the old generic
            self._mod_Delete(dyn_zone, change)
            return

        # IF we're here it's actually an update, sync up rules
        self._mod_dynamic_rulesets(td, change)

    def _mod_dynamic_Delete(self, dyn_zone, change):
        existing = change.existing
        fqdn_tds = self.traffic_directors[existing.fqdn]
        _type = existing._type
        fqdn_tds[_type].delete()
        del fqdn_tds[_type]

    def _mod_Create(self, dyn_zone, change):
        new = change.new
        kwargs_for = getattr(self, '_kwargs_for_{}'.format(new._type))
        for kwargs in kwargs_for(new):
            dyn_zone.add_record(new.name, new._type, **kwargs)

    def _mod_Delete(self, dyn_zone, change):
        existing = change.existing
        if existing.name:
            target = '{}.{}'.format(existing.name, existing.zone.name[:-1])
        else:
            target = existing.zone.name[:-1]
        _type = self.TYPE_TO_RECORDS[existing._type]
        for rec in dyn_zone.get_all_records()[_type]:
            if rec.fqdn == target:
                rec.delete()

    def _mod_Update(self, dyn_zone, change):
        self._mod_Delete(dyn_zone, change)
        self._mod_Create(dyn_zone, change)

    def _apply_traffic_directors(self, desired, changes, dyn_zone):
        self.log.debug('_apply_traffic_directors: zone=%s', desired.name)
        unhandled_changes = []
        for c in changes:
            # we only mess with changes that have geo info somewhere
            if getattr(c.new, 'dynamic', False) or getattr(c.existing,
                                                           'dynamic', False):
                mod = getattr(self, '_mod_dynamic_{}'
                              .format(c.__class__.__name__))
                mod(dyn_zone, c)
            elif getattr(c.new, 'geo', False) or getattr(c.existing, 'geo',
                                                         False):
                mod = getattr(self, '_mod_geo_{}'.format(c.__class__.__name__))
                mod(dyn_zone, c)
            else:
                unhandled_changes.append(c)

        return unhandled_changes

    def _apply_regular(self, desired, changes, dyn_zone):
        self.log.debug('_apply_regular: zone=%s', desired.name)
        for c in changes:
            mod = getattr(self, '_mod_{}'.format(c.__class__.__name__))
            mod(dyn_zone, c)

    # TODO: detect "extra" changes when monitors are out of date or failover
    # chains are wrong etc.

    def _apply(self, plan):
        desired = plan.desired
        changes = plan.changes
        self.log.debug('_apply: zone=%s, len(changes)=%d', desired.name,
                       len(changes))

        self._check_dyn_sess()

        dyn_zone = _CachingDynZone.get(desired.name[:-1], create=True)

        if self.traffic_directors_enabled:
            # any changes left over don't involve geo
            changes = self._apply_traffic_directors(desired, changes, dyn_zone)

        self._apply_regular(desired, changes, dyn_zone)

        dyn_zone.publish()



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from collections import defaultdict
from requests import Session
import logging

from ..record import Record
from .base import BaseProvider


class DigitalOceanClientException(Exception):
    pass


class DigitalOceanClientNotFound(DigitalOceanClientException):

    def __init__(self):
        super(DigitalOceanClientNotFound, self).__init__('Not Found')


class DigitalOceanClientUnauthorized(DigitalOceanClientException):

    def __init__(self):
        super(DigitalOceanClientUnauthorized, self).__init__('Unauthorized')


class DigitalOceanClient(object):
    BASE = 'https://api.digitalocean.com/v2'

    def __init__(self, token):
        sess = Session()
        sess.headers.update({'Authorization': 'Bearer {}'.format(token)})
        self._sess = sess

    def _request(self, method, path, params=None, data=None):
        url = '{}{}'.format(self.BASE, path)
        resp = self._sess.request(method, url, params=params, json=data)
        if resp.status_code == 401:
            raise DigitalOceanClientUnauthorized()
        if resp.status_code == 404:
            raise DigitalOceanClientNotFound()
        resp.raise_for_status()
        return resp

    def domain(self, name):
        path = '/domains/{}'.format(name)
        return self._request('GET', path).json()

    def domain_create(self, name):
        # Digitalocean requires an IP on zone creation
        self._request('POST', '/domains', data={'name': name,
                                                'ip_address': '192.0.2.1'})

        # After the zone is created, immediately delete the record
        records = self.records(name)
        for record in records:
            if record['name'] == '' and record['type'] == 'A':
                self.record_delete(name, record['id'])

    def records(self, zone_name):
        path = '/domains/{}/records'.format(zone_name)
        ret = []

        page = 1
        while True:
            data = self._request('GET', path, {'page': page}).json()

            ret += data['domain_records']
            links = data['links']

            # https://developers.digitalocean.com/documentation/v2/#links
            # pages exists if there is more than 1 page
            # last doesn't exist if you're on the last page
            try:
                links['pages']['last']
                page += 1
            except KeyError:
                break

        for record in ret:
            # change any apex record to empty string
            if record['name'] == '@':
                record['name'] = ''

            # change any apex value to zone name
            if record['data'] == '@':
                record['data'] = zone_name

        return ret

    def record_create(self, zone_name, params):
        path = '/domains/{}/records'.format(zone_name)
        # change empty name string to @, DO uses @ for apex record names
        if params['name'] == '':
            params['name'] = '@'

        self._request('POST', path, data=params)

    def record_delete(self, zone_name, record_id):
        path = '/domains/{}/records/{}'.format(zone_name, record_id)
        self._request('DELETE', path)


class DigitalOceanProvider(BaseProvider):
    '''
    DigitalOcean DNS provider using API v2

    digitalocean:
        class: octodns.provider.digitalocean.DigitalOceanProvider
        # Your DigitalOcean API token (required)
        token: foo
    '''
    SUPPORTS_GEO = False
    SUPPORTS_DYNAMIC = False
    SUPPORTS = set(('A', 'AAAA', 'CAA', 'CNAME', 'MX', 'NS', 'TXT', 'SRV'))

    def __init__(self, id, token, *args, **kwargs):
        self.log = logging.getLogger('DigitalOceanProvider[{}]'.format(id))
        self.log.debug('__init__: id=%s, token=***', id)
        super(DigitalOceanProvider, self).__init__(id, *args, **kwargs)
        self._client = DigitalOceanClient(token)

        self._zone_records = {}

    def _data_for_multiple(self, _type, records):
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': [r['data'] for r in records]
        }

    _data_for_A = _data_for_multiple
    _data_for_AAAA = _data_for_multiple

    def _data_for_CAA(self, _type, records):
        values = []
        for record in records:
            values.append({
                'flags': record['flags'],
                'tag': record['tag'],
                'value': record['data'],
            })
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': values
        }

    def _data_for_CNAME(self, _type, records):
        record = records[0]
        return {
            'ttl': record['ttl'],
            'type': _type,
            'value': '{}.'.format(record['data'])
        }

    def _data_for_MX(self, _type, records):
        values = []
        for record in records:
            values.append({
                'preference': record['priority'],
                'exchange': '{}.'.format(record['data'])
            })
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': values
        }

    def _data_for_NS(self, _type, records):
        values = []
        for record in records:
            data = '{}.'.format(record['data'])
            values.append(data)
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': values,
        }

    def _data_for_SRV(self, _type, records):
        values = []
        for record in records:
            values.append({
                'port': record['port'],
                'priority': record['priority'],
                'target': '{}.'.format(record['data']),
                'weight': record['weight']
            })
        return {
            'type': _type,
            'ttl': records[0]['ttl'],
            'values': values
        }

    def _data_for_TXT(self, _type, records):
        values = [value['data'].replace(';', '\\;') for value in records]
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': values
        }

    def zone_records(self, zone):
        if zone.name not in self._zone_records:
            try:
                self._zone_records[zone.name] = \
                    self._client.records(zone.name[:-1])
            except DigitalOceanClientNotFound:
                return []

        return self._zone_records[zone.name]

    def populate(self, zone, target=False, lenient=False):
        self.log.debug('populate: name=%s, target=%s, lenient=%s', zone.name,
                       target, lenient)

        values = defaultdict(lambda: defaultdict(list))
        for record in self.zone_records(zone):
            _type = record['type']
            values[record['name']][record['type']].append(record)

        before = len(zone.records)
        for name, types in values.items():
            for _type, records in types.items():
                data_for = getattr(self, '_data_for_{}'.format(_type))
                record = Record.new(zone, name, data_for(_type, records),
                                    source=self, lenient=lenient)
                zone.add_record(record, lenient=lenient)

        exists = zone.name in self._zone_records
        self.log.info('populate:   found %s records, exists=%s',
                      len(zone.records) - before, exists)
        return exists

    def _params_for_multiple(self, record):
        for value in record.values:
            yield {
                'data': value,
                'name': record.name,
                'ttl': record.ttl,
                'type': record._type
            }

    _params_for_A = _params_for_multiple
    _params_for_AAAA = _params_for_multiple
    _params_for_NS = _params_for_multiple

    def _params_for_CAA(self, record):
        for value in record.values:
            yield {
                'data': '{}.'.format(value.value),
                'flags': value.flags,
                'name': record.name,
                'tag': value.tag,
                'ttl': record.ttl,
                'type': record._type
            }

    def _params_for_single(self, record):
        yield {
            'data': record.value,
            'name': record.name,
            'ttl': record.ttl,
            'type': record._type
        }

    _params_for_CNAME = _params_for_single

    def _params_for_MX(self, record):
        for value in record.values:
            yield {
                'data': value.exchange,
                'name': record.name,
                'priority': value.preference,
                'ttl': record.ttl,
                'type': record._type
            }

    def _params_for_SRV(self, record):
        for value in record.values:
            yield {
                'data': value.target,
                'name': record.name,
                'port': value.port,
                'priority': value.priority,
                'ttl': record.ttl,
                'type': record._type,
                'weight': value.weight
            }

    def _params_for_TXT(self, record):
        # DigitalOcean doesn't want things escaped in values so we
        # have to strip them here and add them when going the other way
        for value in record.values:
            yield {
                'data': value.replace('\\;', ';'),
                'name': record.name,
                'ttl': record.ttl,
                'type': record._type
            }

    def _apply_Create(self, change):
        new = change.new
        params_for = getattr(self, '_params_for_{}'.format(new._type))
        for params in params_for(new):
            self._client.record_create(new.zone.name[:-1], params)

    def _apply_Update(self, change):
        self._apply_Delete(change)
        self._apply_Create(change)

    def _apply_Delete(self, change):
        existing = change.existing
        zone = existing.zone
        for record in self.zone_records(zone):
            if existing.name == record['name'] and \
               existing._type == record['type']:
                self._client.record_delete(zone.name[:-1], record['id'])

    def _apply(self, plan):
        desired = plan.desired
        changes = plan.changes
        self.log.debug('_apply: zone=%s, len(changes)=%d', desired.name,
                       len(changes))

        domain_name = desired.name[:-1]
        try:
            self._client.domain(domain_name)
        except DigitalOceanClientNotFound:
            self.log.debug('_apply:   no matching zone, creating domain')
            self._client.domain_create(domain_name)

        for change in changes:
            class_name = change.__class__.__name__
            getattr(self, '_apply_{}'.format(class_name))(change)

        # Clear out the cache if any
        self._zone_records.pop(desired.name, None)



#
#
#

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from botocore.exceptions import ClientError
from botocore.stub import ANY, Stubber
from unittest import TestCase
from mock import patch

from octodns.record import Create, Delete, Record, Update
from octodns.provider.route53 import Route53Provider, _Route53GeoDefault, \
    _Route53DynamicValue, _Route53GeoRecord, _Route53Record, _mod_keyer, \
    _octal_replace
from octodns.zone import Zone

from helpers import GeoProvider


class DummyR53Record(object):

    def __init__(self, health_check_id):
        self.health_check_id = health_check_id


class TestOctalReplace(TestCase):

    def test_basic(self):
        for expected, s in (
            ('', ''),
            ('abc', 'abc'),
            ('123', '123'),
            ('abc123', 'abc123'),
            ('*', '\\052'),
            ('abc*', 'abc\\052'),
            ('*abc', '\\052abc'),
            ('123*', '123\\052'),
            ('*123', '\\052123'),
            ('**', '\\052\\052'),
        ):
            self.assertEquals(expected, _octal_replace(s))


dynamic_rrsets = [{
    'Name': '_octodns-default-pool.unit.tests.',
    'ResourceRecords': [{'Value': '1.1.2.1'},
                        {'Value': '1.1.2.2'}],
    'TTL': 60,
    'Type': 'A',
}, {
    'HealthCheckId': '76',
    'Name': '_octodns-ap-southeast-1-value.unit.tests.',
    'ResourceRecords': [{'Value': '1.4.1.1'}],
    'SetIdentifier': 'ap-southeast-1-000',
    'TTL': 60,
    'Type': 'A',
    'Weight': 2
}, {
    'HealthCheckId': '09',
    'Name': '_octodns-ap-southeast-1-value.unit.tests.',
    'ResourceRecords': [{'Value': '1.4.1.2'}],
    'SetIdentifier': 'ap-southeast-1-001',
    'TTL': 60,
    'Type': 'A',
    'Weight': 2
}, {
    'HealthCheckId': 'ab',
    'Name': '_octodns-eu-central-1-value.unit.tests.',
    'ResourceRecords': [{'Value': '1.3.1.1'}],
    'SetIdentifier': 'eu-central-1-000',
    'TTL': 60,
    'Type': 'A',
    'Weight': 1
}, {
    'HealthCheckId': '1e',
    'Name': '_octodns-eu-central-1-value.unit.tests.',
    'ResourceRecords': [{'Value': '1.3.1.2'}],
    'SetIdentifier': 'eu-central-1-001',
    'TTL': 60,
    'Type': 'A',
    'Weight': 1
}, {
    'HealthCheckId': '2a',
    'Name': '_octodns-us-east-1-value.unit.tests.',
    'ResourceRecords': [{'Value': '1.5.1.1'}],
    'SetIdentifier': 'us-east-1-000',
    'TTL': 60,
    'Type': 'A',
    'Weight': 1
}, {
    'HealthCheckId': '61',
    'Name': '_octodns-us-east-1-value.unit.tests.',
    'ResourceRecords': [{'Value': '1.5.1.2'}],
    'SetIdentifier': 'us-east-1-001',
    'TTL': 60,
    'Type': 'A',
    'Weight': 1,
}, {
    'AliasTarget': {'DNSName': '_octodns-default-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'Z2'},
    'Failover': 'SECONDARY',
    'Name': '_octodns-us-east-1-pool.unit.tests.',
    'SetIdentifier': 'us-east-1-Secondary-default',
    'Type': 'A'
}, {
    'AliasTarget': {
        'DNSName': '_octodns-us-east-1-value.unit.tests.',
        'EvaluateTargetHealth': True,
        'HostedZoneId': 'Z2'
    },
    'Failover': 'PRIMARY',
    'Name': '_octodns-us-east-1-pool.unit.tests.',
    'SetIdentifier': 'us-east-1-Primary',
    'Type': 'A',
}, {
    'AliasTarget': {'DNSName': '_octodns-us-east-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'Z2'},
    'Failover': 'SECONDARY',
    'Name': '_octodns-eu-central-1-pool.unit.tests.',
    'SetIdentifier': 'eu-central-1-Secondary-default',
    'Type': 'A'
}, {
    'AliasTarget': {
        'DNSName': '_octodns-eu-central-1-value.unit.tests.',
        'EvaluateTargetHealth': True,
        'HostedZoneId': 'Z2'
    },
    'Failover': 'PRIMARY',
    'Name': '_octodns-eu-central-1-pool.unit.tests.',
    'SetIdentifier': 'eu-central-1-Primary',
    'Type': 'A',
}, {
    'AliasTarget': {'DNSName': '_octodns-us-east-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'Z2'},
    'Failover': 'SECONDARY',
    'Name': '_octodns-ap-southeast-1-pool.unit.tests.',
    'SetIdentifier': 'ap-southeast-1-Secondary-default',
    'Type': 'A'
}, {
    'AliasTarget': {
        'DNSName': '_octodns-ap-southeast-1-value.unit.tests.',
        'EvaluateTargetHealth': True,
        'HostedZoneId': 'Z2'
    },
    'Failover': 'PRIMARY',
    'Name': '_octodns-ap-southeast-1-pool.unit.tests.',
    'SetIdentifier': 'ap-southeast-1-Primary',
    'Type': 'A',
}, {
    'AliasTarget': {'DNSName': '_octodns-ap-southeast-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'Z2'},
    'GeoLocation': {'CountryCode': 'JP'},
    'Name': 'unit.tests.',
    'SetIdentifier': '1-ap-southeast-1-AS-JP',
    'Type': 'A',
}, {
    'AliasTarget': {'DNSName': '_octodns-ap-southeast-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'Z2'},
    'GeoLocation': {'CountryCode': 'CN'},
    'Name': 'unit.tests.',
    'SetIdentifier': '1-ap-southeast-1-AS-CN',
    'Type': 'A',
}, {
    'AliasTarget': {'DNSName': '_octodns-eu-central-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'Z2'},
    'GeoLocation': {'ContinentCode': 'NA-US-FL'},
    'Name': 'unit.tests.',
    'SetIdentifier': '2-eu-central-1-NA-US-FL',
    'Type': 'A',
}, {
    'AliasTarget': {'DNSName': '_octodns-eu-central-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'Z2'},
    'GeoLocation': {'ContinentCode': 'EU'},
    'Name': 'unit.tests.',
    'SetIdentifier': '2-eu-central-1-EU',
    'Type': 'A',
}, {
    'AliasTarget': {'DNSName': '_octodns-us-east-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'Z2'},
    'GeoLocation': {'CountryCode': '*'},
    'Name': 'unit.tests.',
    'SetIdentifier': '3-us-east-1-None',
    'Type': 'A',
}]

dynamic_record_data = {
    'dynamic': {
        'pools': {
            'ap-southeast-1': {
                'fallback': 'us-east-1',
                'values': [{
                    'weight': 2, 'value': '1.4.1.1'
                }, {
                    'weight': 2, 'value': '1.4.1.2'
                }]
            },
            'eu-central-1': {
                'fallback': 'us-east-1',
                'values': [{
                    'weight': 1, 'value': '1.3.1.1'
                }, {
                    'weight': 1, 'value': '1.3.1.2'
                }],
            },
            'us-east-1': {
                'values': [{
                    'weight': 1, 'value': '1.5.1.1'
                }, {
                    'weight': 1, 'value': '1.5.1.2'
                }],
            }
        },
        'rules': [{
            'geos': ['AS-CN', 'AS-JP'],
            'pool': 'ap-southeast-1',
        }, {
            'geos': ['EU', 'NA-US-FL'],
            'pool': 'eu-central-1',
        }, {
            'pool': 'us-east-1',
        }],
    },
    'ttl': 60,
    'type': 'A',
    'values': [
        '1.1.2.1',
        '1.1.2.2',
    ],
}


class TestRoute53Provider(TestCase):
    expected = Zone('unit.tests.', [])
    for name, data in (
        ('simple',
         {'ttl': 60, 'type': 'A', 'values': ['1.2.3.4', '2.2.3.4']}),
        ('',
         {'ttl': 61, 'type': 'A', 'values': ['2.2.3.4', '3.2.3.4'],
          'geo': {
              'AF': ['4.2.3.4'],
              'NA-US': ['5.2.3.4', '6.2.3.4'],
              'NA-US-CA': ['7.2.3.4']}}),
        ('cname', {'ttl': 62, 'type': 'CNAME', 'value': 'unit.tests.'}),
        ('txt', {'ttl': 63, 'type': 'TXT', 'values': ['Hello World!',
                                                      'Goodbye World?']}),
        ('', {'ttl': 64, 'type': 'MX',
              'values': [{
                  'preference': 10,
                  'exchange': 'smtp-1.unit.tests.',
              }, {
                  'preference': 20,
                  'exchange': 'smtp-2.unit.tests.',
              }]}),
        ('naptr', {'ttl': 65, 'type': 'NAPTR',
                   'value': {
                       'order': 10,
                       'preference': 20,
                       'flags': 'U',
                       'service': 'SIP+D2U',
                       'regexp': '!^.*$!sip:info@bar.example.com!',
                       'replacement': '.',
                   }}),
        ('_srv._tcp', {'ttl': 66, 'type': 'SRV', 'value': {
            'priority': 10,
            'weight': 20,
            'port': 30,
            'target': 'cname.unit.tests.'
        }}),
        ('',
         {'ttl': 67, 'type': 'NS', 'values': ['8.2.3.4.', '9.2.3.4.']}),
        ('sub',
         {'ttl': 68, 'type': 'NS', 'values': ['5.2.3.4.', '6.2.3.4.']}),
        ('',
         {'ttl': 69, 'type': 'CAA', 'value': {
             'flags': 0,
             'tag': 'issue',
             'value': 'ca.unit.tests'
         }}),
    ):
        record = Record.new(expected, name, data)
        expected.add_record(record)

    caller_ref = '{}:A:unit.tests.:1324' \
        .format(Route53Provider.HEALTH_CHECK_VERSION)
    health_checks = [{
        'Id': '42',
        'CallerReference': caller_ref,
        'HealthCheckConfig': {
            'Type': 'HTTPS',
            'FullyQualifiedDomainName': 'unit.tests',
            'IPAddress': '4.2.3.4',
            'ResourcePath': '/_dns',
            'Type': 'HTTPS',
            'Port': 443,
            'MeasureLatency': True,
        },
        'HealthCheckVersion': 2,
    }, {
        'Id': 'ignored-also',
        'CallerReference': 'something-else',
        'HealthCheckConfig': {
            'Type': 'HTTPS',
            'FullyQualifiedDomainName': 'unit.tests',
            'IPAddress': '5.2.3.4',
            'ResourcePath': '/_dns',
            'Type': 'HTTPS',
            'Port': 443,
            'MeasureLatency': True,
        },
        'HealthCheckVersion': 42,
    }, {
        'Id': '43',
        'CallerReference': caller_ref,
        'HealthCheckConfig': {
            'Type': 'HTTPS',
            'FullyQualifiedDomainName': 'unit.tests',
            'IPAddress': '5.2.3.4',
            'ResourcePath': '/_dns',
            'Type': 'HTTPS',
            'Port': 443,
            'MeasureLatency': True,
        },
        'HealthCheckVersion': 2,
    }, {
        'Id': '44',
        'CallerReference': caller_ref,
        'HealthCheckConfig': {
            'Type': 'HTTPS',
            'FullyQualifiedDomainName': 'unit.tests',
            'IPAddress': '7.2.3.4',
            'ResourcePath': '/_dns',
            'Type': 'HTTPS',
            'Port': 443,
            'MeasureLatency': True,
        },
        'HealthCheckVersion': 2,
    }, {
        'Id': '45',
        # won't match anything based on type
        'CallerReference': caller_ref.replace(':A:', ':AAAA:'),
        'HealthCheckConfig': {
            'Type': 'HTTPS',
            'FullyQualifiedDomainName': 'unit.tests',
            'IPAddress': '7.2.3.4',
            'ResourcePath': '/_dns',
            'Type': 'HTTPS',
            'Port': 443,
            'MeasureLatency': True,
        },
        'HealthCheckVersion': 2,
    }]

    def _get_stubbed_provider(self):
        provider = Route53Provider('test', 'abc', '123')

        # Use the stubber
        stubber = Stubber(provider._conn)
        stubber.activate()

        return (provider, stubber)

    def _get_stubbed_fallback_auth_provider(self):
        provider = Route53Provider('test')

        # Use the stubber
        stubber = Stubber(provider._conn)
        stubber.activate()

        return (provider, stubber)

    def test_populate_with_fallback(self):
        provider, stubber = self._get_stubbed_fallback_auth_provider()

        got = Zone('unit.tests.', [])
        with self.assertRaises(ClientError):
            stubber.add_client_error('list_hosted_zones')
            provider.populate(got)

    def test_populate(self):
        provider, stubber = self._get_stubbed_provider()

        got = Zone('unit.tests.', [])
        with self.assertRaises(ClientError):
            stubber.add_client_error('list_hosted_zones')
            provider.populate(got)

        with self.assertRaises(ClientError):
            list_hosted_zones_resp = {
                'HostedZones': [{
                    'Name': 'unit.tests.',
                    'Id': 'z42',
                    'CallerReference': 'abc',
                }],
                'Marker': 'm',
                'IsTruncated': False,
                'MaxItems': '100',
            }
            stubber.add_response('list_hosted_zones', list_hosted_zones_resp,
                                 {})
            stubber.add_client_error('list_resource_record_sets',
                                     expected_params={'HostedZoneId': u'z42'})
            provider.populate(got)
            stubber.assert_no_pending_responses()

        # list_hosted_zones has been cached from now on so we don't have to
        # worry about stubbing it

        list_resource_record_sets_resp_p1 = {
            'ResourceRecordSets': [{
                'Name': 'simple.unit.tests.',
                'Type': 'A',
                'ResourceRecords': [{
                    'Value': '1.2.3.4',
                }, {
                    'Value': '2.2.3.4',
                }],
                'TTL': 60,
            }, {
                'Name': 'unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'CountryCode': '*',
                },
                'ResourceRecords': [{
                    'Value': '2.2.3.4',
                }, {
                    'Value': '3.2.3.4',
                }],
                'TTL': 61,
            }, {
                'Name': 'unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'ContinentCode': 'AF',
                },
                'ResourceRecords': [{
                    'Value': '4.2.3.4',
                }],
                'TTL': 61,
            }, {
                'Name': 'unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'CountryCode': 'US',
                },
                'ResourceRecords': [{
                    'Value': '5.2.3.4',
                }, {
                    'Value': '6.2.3.4',
                }],
                'TTL': 61,
            }, {
                'Name': 'unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'CountryCode': 'US',
                    'SubdivisionCode': 'CA',
                },
                'ResourceRecords': [{
                    'Value': '7.2.3.4',
                }],
                'TTL': 61,
            }],
            'IsTruncated': True,
            'NextRecordName': 'next_name',
            'NextRecordType': 'next_type',
            'MaxItems': '100',
        }
        stubber.add_response('list_resource_record_sets',
                             list_resource_record_sets_resp_p1,
                             {'HostedZoneId': 'z42'})

        list_resource_record_sets_resp_p2 = {
            'ResourceRecordSets': [{
                'Name': 'cname.unit.tests.',
                'Type': 'CNAME',
                'ResourceRecords': [{
                    'Value': 'unit.tests.',
                }],
                'TTL': 62,
            }, {
                'Name': 'txt.unit.tests.',
                'Type': 'TXT',
                'ResourceRecords': [{
                    'Value': '"Hello World!"',
                }, {
                    'Value': '"Goodbye World?"',
                }],
                'TTL': 63,
            }, {
                'Name': 'unit.tests.',
                'Type': 'MX',
                'ResourceRecords': [{
                    'Value': '10 smtp-1.unit.tests.',
                }, {
                    'Value': '20 smtp-2.unit.tests.',
                }],
                'TTL': 64,
            }, {
                'Name': 'naptr.unit.tests.',
                'Type': 'NAPTR',
                'ResourceRecords': [{
                    'Value': '10 20 "U" "SIP+D2U" '
                        '"!^.*$!sip:info@bar.example.com!" .',
                }],
                'TTL': 65,
            }, {
                'Name': '_srv._tcp.unit.tests.',
                'Type': 'SRV',
                'ResourceRecords': [{
                    'Value': '10 20 30 cname.unit.tests.',
                }],
                'TTL': 66,
            }, {
                'Name': 'unit.tests.',
                'Type': 'NS',
                'ResourceRecords': [{
                    'Value': 'ns1.unit.tests.',
                }],
                'TTL': 67,
            }, {
                'Name': 'sub.unit.tests.',
                'Type': 'NS',
                'GeoLocation': {
                    'ContinentCode': 'AF',
                },
                'ResourceRecords': [{
                    'Value': '5.2.3.4.',
                }, {
                    'Value': '6.2.3.4.',
                }],
                'TTL': 68,
            }, {
                'Name': 'soa.unit.tests.',
                'Type': 'SOA',
                'ResourceRecords': [{
                    'Value': 'ns1.unit.tests.',
                }],
                'TTL': 69,
            }, {
                'Name': 'unit.tests.',
                'Type': 'CAA',
                'ResourceRecords': [{
                    'Value': '0 issue "ca.unit.tests"',
                }],
                'TTL': 69,
            }, {
                'AliasTarget': {
                    'HostedZoneId': 'Z119WBBTVP5WFX',
                    'EvaluateTargetHealth': False,
                    'DNSName': 'unit.tests.'
                },
                'Type': 'A',
                'Name': 'alias.unit.tests.'
            }],
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_resource_record_sets',
                             list_resource_record_sets_resp_p2,
                             {'HostedZoneId': 'z42',
                              'StartRecordName': 'next_name',
                              'StartRecordType': 'next_type'})

        # Load everything
        provider.populate(got)
        # Make sure we got what we expected
        changes = self.expected.changes(got, GeoProvider())
        self.assertEquals(0, len(changes))
        stubber.assert_no_pending_responses()

        # Populate a zone that doesn't exist
        nonexistent = Zone('does.not.exist.', [])
        provider.populate(nonexistent)
        self.assertEquals(set(), nonexistent.records)

    def test_sync(self):
        provider, stubber = self._get_stubbed_provider()

        list_hosted_zones_resp = {
            'HostedZones': [{
                'Name': 'unit.tests.',
                'Id': 'z42',
                'CallerReference': 'abc',
            }],
            'Marker': 'm',
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_hosted_zones', list_hosted_zones_resp,
                             {})
        list_resource_record_sets_resp = {
            'ResourceRecordSets': [],
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_resource_record_sets',
                             list_resource_record_sets_resp,
                             {'HostedZoneId': 'z42'})

        plan = provider.plan(self.expected)
        self.assertEquals(9, len(plan.changes))
        self.assertTrue(plan.exists)
        for change in plan.changes:
            self.assertIsInstance(change, Create)
        stubber.assert_no_pending_responses()

        stubber.add_response('list_health_checks',
                             {
                                 'HealthChecks': self.health_checks,
                                 'IsTruncated': False,
                                 'MaxItems': '100',
                                 'Marker': '',
                             })
        stubber.add_response('change_resource_record_sets',
                             {'ChangeInfo': {
                                 'Id': 'id',
                                 'Status': 'PENDING',
                                 'SubmittedAt': '2017-01-29T01:02:03Z',
                             }}, {'HostedZoneId': 'z42', 'ChangeBatch': ANY})

        self.assertEquals(9, provider.apply(plan))
        stubber.assert_no_pending_responses()

        # Delete by monkey patching in a populate that includes an extra record
        def add_extra_populate(existing, target, lenient):
            for record in self.expected.records:
                existing.add_record(record)
            record = Record.new(existing, 'extra',
                                {'ttl': 99, 'type': 'A',
                                 'values': ['9.9.9.9']})
            existing.add_record(record)

        provider.populate = add_extra_populate
        change_resource_record_sets_params = {
            'ChangeBatch': {
                'Changes': [{
                    'Action': 'DELETE', 'ResourceRecordSet': {
                        'Name': 'extra.unit.tests.',
                        'ResourceRecords': [{'Value': u'9.9.9.9'}],
                        'TTL': 99,
                        'Type': 'A'
                    }}],
                u'Comment': ANY
            },
            'HostedZoneId': u'z42'
        }
        stubber.add_response('change_resource_record_sets',
                             {'ChangeInfo': {
                                 'Id': 'id',
                                 'Status': 'PENDING',
                                 'SubmittedAt': '2017-01-29T01:02:03Z',
                             }}, change_resource_record_sets_params)
        plan = provider.plan(self.expected)
        self.assertEquals(1, len(plan.changes))
        self.assertIsInstance(plan.changes[0], Delete)
        self.assertEquals(1, provider.apply(plan))
        stubber.assert_no_pending_responses()

        # Update by monkey patching in a populate that modifies the A record
        # with geos
        def mod_geo_populate(existing, target, lenient):
            for record in self.expected.records:
                if record._type != 'A' or not record.geo:
                    existing.add_record(record)
            record = Record.new(existing, '', {
                'ttl': 61,
                'type': 'A',
                'values': ['8.2.3.4', '3.2.3.4'],
                'geo': {
                    'AF': ['4.2.3.4'],
                    'NA-US': ['5.2.3.4', '6.2.3.4'],
                    'NA-US-KY': ['7.2.3.4']
                }
            })
            existing.add_record(record)

        provider.populate = mod_geo_populate
        change_resource_record_sets_params = {
            'ChangeBatch': {
                'Changes': [{
                    'Action': 'DELETE',
                    'ResourceRecordSet': {
                        'GeoLocation': {'CountryCode': 'US',
                                        'SubdivisionCode': 'KY'},
                        'HealthCheckId': u'44',
                        'Name': 'unit.tests.',
                        'ResourceRecords': [{'Value': '7.2.3.4'}],
                        'SetIdentifier': 'NA-US-KY',
                        'TTL': 61,
                        'Type': 'A'
                    }
                }, {
                    'Action': 'CREATE',
                    'ResourceRecordSet': {
                        'GeoLocation': {'CountryCode': 'US',
                                        'SubdivisionCode': 'CA'},
                        'HealthCheckId': u'44',
                        'Name': 'unit.tests.',
                        'ResourceRecords': [{'Value': '7.2.3.4'}],
                        'SetIdentifier': 'NA-US-CA',
                        'TTL': 61,
                        'Type': 'A'
                    }
                }, {
                    'Action': 'UPSERT',
                    'ResourceRecordSet': {
                        'GeoLocation': {'ContinentCode': 'AF'},
                        'Name': 'unit.tests.',
                        'HealthCheckId': u'42',
                        'ResourceRecords': [{'Value': '4.2.3.4'}],
                        'SetIdentifier': 'AF',
                        'TTL': 61,
                        'Type': 'A'
                    }
                }, {
                    'Action': 'UPSERT',
                    'ResourceRecordSet': {
                        'GeoLocation': {'CountryCode': 'US'},
                        'HealthCheckId': u'43',
                        'Name': 'unit.tests.',
                        'ResourceRecords': [{'Value': '5.2.3.4'},
                                            {'Value': '6.2.3.4'}],
                        'SetIdentifier': 'NA-US',
                        'TTL': 61,
                        'Type': 'A'
                    }
                }, {
                    'Action': 'UPSERT',
                    'ResourceRecordSet': {
                        'GeoLocation': {'CountryCode': '*'},
                        'Name': 'unit.tests.',
                        'ResourceRecords': [{'Value': '2.2.3.4'},
                                            {'Value': '3.2.3.4'}],
                        'SetIdentifier': 'default',
                        'TTL': 61,
                        'Type': 'A'
                    }
                }],
                'Comment': ANY
            },
            'HostedZoneId': 'z42'
        }
        stubber.add_response('change_resource_record_sets',
                             {'ChangeInfo': {
                                 'Id': 'id',
                                 'Status': 'PENDING',
                                 'SubmittedAt': '2017-01-29T01:02:03Z',
                             }}, change_resource_record_sets_params)
        plan = provider.plan(self.expected)
        self.assertEquals(1, len(plan.changes))
        self.assertIsInstance(plan.changes[0], Update)
        self.assertEquals(1, provider.apply(plan))
        stubber.assert_no_pending_responses()

        # Update converting to non-geo by monkey patching in a populate that
        # modifies the A record with geos
        def mod_add_geo_populate(existing, target, lenient):
            for record in self.expected.records:
                if record._type != 'A' or record.geo:
                    existing.add_record(record)
            record = Record.new(existing, 'simple', {
                'ttl': 61,
                'type': 'A',
                'values': ['1.2.3.4', '2.2.3.4'],
                'geo': {
                    'OC': ['3.2.3.4', '4.2.3.4'],
                }
            })
            existing.add_record(record)

        provider.populate = mod_add_geo_populate
        change_resource_record_sets_params = {
            'ChangeBatch': {
                'Changes': [{
                    'Action': 'DELETE',
                    'ResourceRecordSet': {
                        'GeoLocation': {'ContinentCode': 'OC'},
                        'Name': 'simple.unit.tests.',
                        'ResourceRecords': [{'Value': '3.2.3.4'},
                                            {'Value': '4.2.3.4'}],
                        'SetIdentifier': 'OC',
                        'TTL': 61,
                        'Type': 'A'}
                }, {
                    'Action': 'DELETE',
                    'ResourceRecordSet': {
                        'GeoLocation': {'CountryCode': '*'},
                        'Name': 'simple.unit.tests.',
                        'ResourceRecords': [{'Value': '1.2.3.4'},
                                            {'Value': '2.2.3.4'}],
                        'SetIdentifier': 'default',
                        'TTL': 61,
                        'Type': 'A'}
                }, {
                    'Action': 'CREATE',
                    'ResourceRecordSet': {
                        'Name': 'simple.unit.tests.',
                        'ResourceRecords': [{'Value': '1.2.3.4'},
                                            {'Value': '2.2.3.4'}],
                        'TTL': 60,
                        'Type': 'A'}
                }],
                'Comment': ANY
            },
            'HostedZoneId': 'z42'
        }
        stubber.add_response('change_resource_record_sets',
                             {'ChangeInfo': {
                                 'Id': 'id',
                                 'Status': 'PENDING',
                                 'SubmittedAt': '2017-01-29T01:02:03Z',
                             }}, change_resource_record_sets_params)
        plan = provider.plan(self.expected)
        self.assertEquals(1, len(plan.changes))
        self.assertIsInstance(plan.changes[0], Update)
        self.assertEquals(1, provider.apply(plan))
        stubber.assert_no_pending_responses()

    def test_sync_create(self):
        provider, stubber = self._get_stubbed_provider()

        got = Zone('unit.tests.', [])

        list_hosted_zones_resp = {
            'HostedZones': [],
            'Marker': 'm',
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_hosted_zones', list_hosted_zones_resp,
                             {})

        plan = provider.plan(self.expected)
        self.assertEquals(9, len(plan.changes))
        self.assertFalse(plan.exists)
        for change in plan.changes:
            self.assertIsInstance(change, Create)
        stubber.assert_no_pending_responses()

        create_hosted_zone_resp = {
            'HostedZone': {
                'Name': 'unit.tests.',
                'Id': 'z42',
                'CallerReference': 'abc',
            },
            'ChangeInfo': {
                'Id': 'a12',
                'Status': 'PENDING',
                'SubmittedAt': '2017-01-29T01:02:03Z',
                'Comment': 'hrm',
            },
            'DelegationSet': {
                'Id': 'b23',
                'CallerReference': 'blip',
                'NameServers': [
                    'n12.unit.tests.',
                ],
            },
            'Location': 'us-east-1',
        }
        stubber.add_response('create_hosted_zone',
                             create_hosted_zone_resp, {
                                 'Name': got.name,
                                 'CallerReference': ANY,
                             })

        list_resource_record_sets_resp = {
            'ResourceRecordSets': [{
                'Name': 'a.unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'ContinentCode': 'NA',
                },
                'ResourceRecords': [{
                    'Value': '2.2.3.4',
                }],
                'TTL': 61,
            }],
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_resource_record_sets',
                             list_resource_record_sets_resp,
                             {'HostedZoneId': 'z42'})

        stubber.add_response('list_health_checks',
                             {
                                 'HealthChecks': self.health_checks,
                                 'IsTruncated': False,
                                 'MaxItems': '100',
                                 'Marker': '',
                             })

        stubber.add_response('change_resource_record_sets',
                             {'ChangeInfo': {
                                 'Id': 'id',
                                 'Status': 'PENDING',
                                 'SubmittedAt': '2017-01-29T01:02:03Z',
                             }}, {'HostedZoneId': 'z42', 'ChangeBatch': ANY})

        self.assertEquals(9, provider.apply(plan))
        stubber.assert_no_pending_responses()

    def test_health_checks_pagination(self):
        provider, stubber = self._get_stubbed_provider()

        health_checks_p1 = [{
            'Id': '42',
            'CallerReference': self.caller_ref,
            'HealthCheckConfig': {
                'Type': 'HTTPS',
                'FullyQualifiedDomainName': 'unit.tests',
                'IPAddress': '4.2.3.4',
                'ResourcePath': '/_dns',
                'Type': 'HTTPS',
                'Port': 443,
                'MeasureLatency': True,
            },
            'HealthCheckVersion': 2,
        }, {
            'Id': '43',
            'CallerReference': 'abc123',
            'HealthCheckConfig': {
                'Type': 'HTTPS',
                'FullyQualifiedDomainName': 'unit.tests',
                'IPAddress': '9.2.3.4',
                'ResourcePath': '/_dns',
                'Type': 'HTTPS',
                'Port': 443,
                'MeasureLatency': True,
            },
            'HealthCheckVersion': 2,
        }]
        stubber.add_response('list_health_checks',
                             {
                                 'HealthChecks': health_checks_p1,
                                 'IsTruncated': True,
                                 'MaxItems': '2',
                                 'Marker': '',
                                 'NextMarker': 'moar',
                             })

        health_checks_p2 = [{
            'Id': '44',
            'CallerReference': self.caller_ref,
            'HealthCheckConfig': {
                'Type': 'HTTPS',
                'FullyQualifiedDomainName': 'unit.tests',
                'IPAddress': '8.2.3.4',
                'ResourcePath': '/_dns',
                'Type': 'HTTPS',
                'Port': 443,
                'MeasureLatency': True,
            },
            'HealthCheckVersion': 2,
        }]
        stubber.add_response('list_health_checks',
                             {
                                 'HealthChecks': health_checks_p2,
                                 'IsTruncated': False,
                                 'MaxItems': '2',
                                 'Marker': 'moar',
                             }, {'Marker': 'moar'})

        health_checks = provider.health_checks
        self.assertEquals({
            '42': health_checks_p1[0],
            '44': health_checks_p2[0],
        }, health_checks)
        stubber.assert_no_pending_responses()

        # get without create
        record = Record.new(self.expected, '', {
            'ttl': 61,
            'type': 'A',
            'values': ['2.2.3.4', '3.2.3.4'],
            'geo': {
                'AF': ['4.2.3.4'],
            }
        })
        value = record.geo['AF'].values[0]
        id = provider.get_health_check_id(record, value, True)
        self.assertEquals('42', id)

    def test_health_check_create(self):
        provider, stubber = self._get_stubbed_provider()

        # No match based on type
        caller_ref = \
            '{}:AAAA:foo1234'.format(Route53Provider.HEALTH_CHECK_VERSION)
        health_checks = [{
            'Id': '42',
            # No match based on version
            'CallerReference': '9999:A:foo1234',
            'HealthCheckConfig': {
                'Type': 'HTTPS',
                'FullyQualifiedDomainName': 'unit.tests',
                'IPAddress': '4.2.3.4',
                'ResourcePath': '/_dns',
                'Type': 'HTTPS',
                'Port': 443,
                'MeasureLatency': True,
            },
            'HealthCheckVersion': 2,
        }, {
            'Id': '43',
            'CallerReference': caller_ref,
            'HealthCheckConfig': {
                'Type': 'HTTPS',
                'FullyQualifiedDomainName': 'unit.tests',
                'IPAddress': '4.2.3.4',
                'ResourcePath': '/_dns',
                'Type': 'HTTPS',
                'Port': 443,
                'MeasureLatency': True,
            },
            'HealthCheckVersion': 2,
        }]
        stubber.add_response('list_health_checks', {
            'HealthChecks': health_checks,
            'IsTruncated': False,
            'MaxItems': '100',
            'Marker': '',
        })

        health_check_config = {
            'EnableSNI': False,
            'FailureThreshold': 6,
            'FullyQualifiedDomainName': 'foo.bar.com',
            'IPAddress': '4.2.3.4',
            'MeasureLatency': True,
            'Port': 8080,
            'RequestInterval': 10,
            'ResourcePath': '/_status',
            'Type': 'HTTP'
        }
        stubber.add_response('create_health_check', {
            'HealthCheck': {
                'Id': '42',
                'CallerReference': self.caller_ref,
                'HealthCheckConfig': health_check_config,
                'HealthCheckVersion': 1,
            },
            'Location': 'http://url',
        }, {
            'CallerReference': ANY,
            'HealthCheckConfig': health_check_config,
        })
        stubber.add_response('change_tags_for_resource', {})

        record = Record.new(self.expected, '', {
            'ttl': 61,
            'type': 'A',
            'values': ['2.2.3.4', '3.2.3.4'],
            'geo': {
                'AF': ['4.2.3.4'],
            },
            'octodns': {
                'healthcheck': {
                    'host': 'foo.bar.com',
                    'path': '/_status',
                    'port': 8080,
                    'protocol': 'HTTP',
                },
            }
        })

        # if not allowed to create returns none
        value = record.geo['AF'].values[0]
        id = provider.get_health_check_id(record, value, False)
        self.assertFalse(id)

        # when allowed to create we do
        id = provider.get_health_check_id(record, value, True)
        self.assertEquals('42', id)
        stubber.assert_no_pending_responses()

        # A CNAME style healthcheck, without a value

        health_check_config = {
            'EnableSNI': False,
            'FailureThreshold': 6,
            'FullyQualifiedDomainName': 'target-1.unit.tests.',
            'MeasureLatency': True,
            'Port': 8080,
            'RequestInterval': 10,
            'ResourcePath': '/_status',
            'Type': 'HTTP'
        }
        stubber.add_response('create_health_check', {
            'HealthCheck': {
                'Id': '42',
                'CallerReference': self.caller_ref,
                'HealthCheckConfig': health_check_config,
                'HealthCheckVersion': 1,
            },
            'Location': 'http://url',
        }, {
            'CallerReference': ANY,
            'HealthCheckConfig': health_check_config,
        })
        stubber.add_response('change_tags_for_resource', {})

        id = provider.get_health_check_id(record, 'target-1.unit.tests.', True)
        self.assertEquals('42', id)
        stubber.assert_no_pending_responses()

    def test_health_check_measure_latency(self):
        provider, stubber = self._get_stubbed_provider()
        record_true = Record.new(self.expected, 'a', {
            'ttl': 61,
            'type': 'A',
            'value': '1.2.3.4',
            'octodns': {
                'healthcheck': {
                },
                'route53': {
                    'healthcheck': {
                        'measure_latency': True
                    }
                }
            }
        })
        measure_latency = provider._healthcheck_measure_latency(record_true)
        self.assertTrue(measure_latency)

        record_default = Record.new(self.expected, 'a', {
            'ttl': 61,
            'type': 'A',
            'value': '1.2.3.4',
        })
        measure_latency = provider._healthcheck_measure_latency(record_default)
        self.assertTrue(measure_latency)

        record_false = Record.new(self.expected, 'a', {
            'ttl': 61,
            'type': 'A',
            'value': '1.2.3.4',
            'octodns': {
                'healthcheck': {
                },
                'route53': {
                    'healthcheck': {
                        'measure_latency': False
                    }
                }
            }
        })
        measure_latency = provider._healthcheck_measure_latency(record_false)
        self.assertFalse(measure_latency)

    def test_create_health_checks_measure_latency(self):
        provider, stubber = self._get_stubbed_provider()

        health_check_config = {
            'EnableSNI': True,
            'FailureThreshold': 6,
            'FullyQualifiedDomainName': 'a.unit.tests',
            'IPAddress': '1.2.3.4',
            'MeasureLatency': False,
            'Port': 443,
            'RequestInterval': 10,
            'ResourcePath': '/_dns',
            'Type': 'HTTPS'
        }

        stubber.add_response('list_health_checks', {
            'HealthChecks': [],
            'IsTruncated': False,
            'MaxItems': '100',
            'Marker': '',
        })

        stubber.add_response('create_health_check', {
            'HealthCheck': {
                'Id': '42',
                'CallerReference': self.caller_ref,
                'HealthCheckConfig': health_check_config,
                'HealthCheckVersion': 1,
            },
            'Location': 'http://url',
        }, {
            'CallerReference': ANY,
            'HealthCheckConfig': health_check_config,
        })
        stubber.add_response('change_tags_for_resource', {})
        stubber.add_response('change_tags_for_resource', {})

        record = Record.new(self.expected, 'a', {
            'ttl': 61,
            'type': 'A',
            'value': '2.2.3.4',
            'geo': {
                'AF': ['1.2.3.4'],
            },
            'octodns': {
                'healthcheck': {
                },
                'route53': {
                    'healthcheck': {
                        'measure_latency': False
                    }
                }
            }
        })

        value = record.geo['AF'].values[0]
        id = provider.get_health_check_id(record, value, True)
        ml = provider.health_checks[id]['HealthCheckConfig']['MeasureLatency']
        self.assertEqual(False, ml)

    def test_health_check_gc(self):
        provider, stubber = self._get_stubbed_provider()

        stubber.add_response('list_health_checks', {
            'HealthChecks': self.health_checks,
            'IsTruncated': False,
            'MaxItems': '100',
            'Marker': '',
        })

        record = Record.new(self.expected, '', {
            'ttl': 61,
            'type': 'A',
            'values': ['2.2.3.4', '3.2.3.4'],
            'geo': {
                'AF': ['4.2.3.4'],
                'NA-US': ['5.2.3.4', '6.2.3.4'],
                # removed one geo
            }
        })

        # gc no longer in_use records (directly)
        stubber.add_response('delete_health_check', {}, {
            'HealthCheckId': '44',
        })
        provider._gc_health_checks(record, [
            DummyR53Record('42'),
            DummyR53Record('43'),
        ])
        stubber.assert_no_pending_responses()

        # gc through _mod_Create
        stubber.add_response('delete_health_check', {}, {
            'HealthCheckId': '44',
        })
        change = Create(record)
        provider._mod_Create(change, 'z43', [])
        stubber.assert_no_pending_responses()

        # gc through _mod_Update
        stubber.add_response('delete_health_check', {}, {
            'HealthCheckId': '44',
        })
        # first record is ignored for our purposes, we have to pass something
        change = Update(record, record)
        provider._mod_Create(change, 'z43', [])
        stubber.assert_no_pending_responses()

        # gc through _mod_Delete, expect 3 to go away, can't check order
        # b/c it's not deterministic
        stubber.add_response('delete_health_check', {}, {
            'HealthCheckId': ANY,
        })
        stubber.add_response('delete_health_check', {}, {
            'HealthCheckId': ANY,
        })
        stubber.add_response('delete_health_check', {}, {
            'HealthCheckId': ANY,
        })
        change = Delete(record)
        provider._mod_Delete(change, 'z43', [])
        stubber.assert_no_pending_responses()

        # gc only AAAA, leave the A's alone
        stubber.add_response('delete_health_check', {}, {
            'HealthCheckId': '45',
        })
        record = Record.new(self.expected, '', {
            'ttl': 61,
            'type': 'AAAA',
            'value': '2001:0db8:3c4d:0015:0000:0000:1a2f:1a4b'
        })
        provider._gc_health_checks(record, [])
        stubber.assert_no_pending_responses()

    def test_legacy_health_check_gc(self):
        provider, stubber = self._get_stubbed_provider()

        old_caller_ref = '0000:A:3333'
        health_checks = [{
            'Id': '42',
            'CallerReference': self.caller_ref,
            'HealthCheckConfig': {
                'Type': 'HTTPS',
                'FullyQualifiedDomainName': 'unit.tests',
                'IPAddress': '4.2.3.4',
                'ResourcePath': '/_dns',
                'Type': 'HTTPS',
                'Port': 443,
                'MeasureLatency': True,
            },
            'HealthCheckVersion': 2,
        }, {
            'Id': '43',
            'CallerReference': old_caller_ref,
            'HealthCheckConfig': {
                'Type': 'HTTPS',
                'FullyQualifiedDomainName': 'unit.tests',
                'IPAddress': '4.2.3.4',
                'ResourcePath': '/_dns',
                'Type': 'HTTPS',
                'Port': 443,
                'MeasureLatency': True,
            },
            'HealthCheckVersion': 2,
        }, {
            'Id': '44',
            'CallerReference': old_caller_ref,
            'HealthCheckConfig': {
                'Type': 'HTTPS',
                'FullyQualifiedDomainName': 'other.unit.tests',
                'IPAddress': '4.2.3.4',
                'ResourcePath': '/_dns',
                'Type': 'HTTPS',
                'Port': 443,
                'MeasureLatency': True,
            },
            'HealthCheckVersion': 2,
        }]

        stubber.add_response('list_health_checks', {
            'HealthChecks': health_checks,
            'IsTruncated': False,
            'MaxItems': '100',
            'Marker': '',
        })

        # No changes to the record itself
        record = Record.new(self.expected, '', {
            'ttl': 61,
            'type': 'A',
            'values': ['2.2.3.4', '3.2.3.4'],
            'geo': {
                'AF': ['4.2.3.4'],
                'NA-US': ['5.2.3.4', '6.2.3.4'],
                'NA-US-CA': ['7.2.3.4']
            }
        })

        # Expect to delete the legacy hc for our record, but not touch the new
        # one or the other legacy record
        stubber.add_response('delete_health_check', {}, {
            'HealthCheckId': '43',
        })

        provider._gc_health_checks(record, [
            DummyR53Record('42'),
        ])

    def test_no_extra_changes(self):
        provider, stubber = self._get_stubbed_provider()

        list_hosted_zones_resp = {
            'HostedZones': [{
                'Name': 'unit.tests.',
                'Id': 'z42',
                'CallerReference': 'abc',
            }],
            'Marker': 'm',
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_hosted_zones', list_hosted_zones_resp, {})

        # empty is empty
        desired = Zone('unit.tests.', [])
        extra = provider._extra_changes(desired=desired, changes=[])
        self.assertEquals([], extra)
        stubber.assert_no_pending_responses()

        # single record w/o geo is empty
        desired = Zone('unit.tests.', [])
        record = Record.new(desired, 'a', {
            'ttl': 30,
            'type': 'A',
            'value': '1.2.3.4',
        })
        desired.add_record(record)
        extra = provider._extra_changes(desired=desired, changes=[])
        self.assertEquals([], extra)
        stubber.assert_no_pending_responses()

        # short-circuit for unknown zone
        other = Zone('other.tests.', [])
        extra = provider._extra_changes(desired=other, changes=[])
        self.assertEquals([], extra)
        stubber.assert_no_pending_responses()

    def test_extra_change_no_health_check(self):
        provider, stubber = self._get_stubbed_provider()

        list_hosted_zones_resp = {
            'HostedZones': [{
                'Name': 'unit.tests.',
                'Id': 'z42',
                'CallerReference': 'abc',
            }],
            'Marker': 'm',
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_hosted_zones', list_hosted_zones_resp, {})

        # record with geo and no health check returns change
        desired = Zone('unit.tests.', [])
        record = Record.new(desired, 'a', {
            'ttl': 30,
            'type': 'A',
            'value': '1.2.3.4',
            'geo': {
                'NA': ['2.2.3.4'],
            }
        })
        desired.add_record(record)
        list_resource_record_sets_resp = {
            'ResourceRecordSets': [{
                'Name': 'a.unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'ContinentCode': 'NA',
                },
                'ResourceRecords': [{
                    'Value': '2.2.3.4',
                }],
                'TTL': 61,
            }],
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_resource_record_sets',
                             list_resource_record_sets_resp,
                             {'HostedZoneId': 'z42'})
        extra = provider._extra_changes(desired=desired, changes=[])
        self.assertEquals(1, len(extra))
        stubber.assert_no_pending_responses()

    def test_extra_change_has_wrong_health_check(self):
        provider, stubber = self._get_stubbed_provider()

        list_hosted_zones_resp = {
            'HostedZones': [{
                'Name': 'unit.tests.',
                'Id': 'z42',
                'CallerReference': 'abc',
            }],
            'Marker': 'm',
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_hosted_zones', list_hosted_zones_resp, {})

        # record with geo and no health check returns change
        desired = Zone('unit.tests.', [])
        record = Record.new(desired, 'a', {
            'ttl': 30,
            'type': 'A',
            'value': '1.2.3.4',
            'geo': {
                'NA': ['2.2.3.4'],
            }
        })
        desired.add_record(record)
        list_resource_record_sets_resp = {
            'ResourceRecordSets': [{
                'Name': 'a.unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'ContinentCode': 'NA',
                },
                'ResourceRecords': [{
                    'Value': '2.2.3.4',
                }],
                'TTL': 61,
                'HealthCheckId': '42',
            }],
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_resource_record_sets',
                             list_resource_record_sets_resp,
                             {'HostedZoneId': 'z42'})
        stubber.add_response('list_health_checks', {
            'HealthChecks': [{
                'Id': '42',
                'CallerReference': 'foo',
                'HealthCheckConfig': {
                    'Type': 'HTTPS',
                    'FullyQualifiedDomainName': 'unit.tests',
                    'IPAddress': '2.2.3.4',
                    'ResourcePath': '/_dns',
                    'Type': 'HTTPS',
                    'Port': 443,
                    'MeasureLatency': True,
                },
                'HealthCheckVersion': 2,
            }],
            'IsTruncated': False,
            'MaxItems': '100',
            'Marker': '',
        })
        extra = provider._extra_changes(desired=desired, changes=[])
        self.assertEquals(1, len(extra))
        stubber.assert_no_pending_responses()

        for change in (Create(record), Update(record, record), Delete(record)):
            extra = provider._extra_changes(desired=desired, changes=[change])
            self.assertEquals(0, len(extra))
            stubber.assert_no_pending_responses()

    def test_extra_change_has_health_check(self):
        provider, stubber = self._get_stubbed_provider()

        list_hosted_zones_resp = {
            'HostedZones': [{
                'Name': 'unit.tests.',
                'Id': 'z42',
                'CallerReference': 'abc',
            }],
            'Marker': 'm',
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_hosted_zones', list_hosted_zones_resp, {})

        # record with geo and no health check returns change
        desired = Zone('unit.tests.', [])
        record = Record.new(desired, 'a', {
            'ttl': 30,
            'type': 'A',
            'value': '1.2.3.4',
            'geo': {
                'NA': ['2.2.3.4'],
            }
        })
        desired.add_record(record)
        list_resource_record_sets_resp = {
            'ResourceRecordSets': [{
                # other name
                'Name': 'unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'CountryCode': '*',
                },
                'ResourceRecords': [{
                    'Value': '1.2.3.4',
                }],
                'TTL': 61,
            }, {
                # matching name, other type
                'Name': 'a.unit.tests.',
                'Type': 'AAAA',
                'ResourceRecords': [{
                    'Value': '2001:0db8:3c4d:0015:0000:0000:1a2f:1a4b'
                }],
                'TTL': 61,
            }, {
                # default geo
                'Name': 'a.unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'CountryCode': '*',
                },
                'ResourceRecords': [{
                    'Value': '1.2.3.4',
                }],
                'TTL': 61,
            }, {
                # match w/correct geo
                'Name': 'a.unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'ContinentCode': 'NA',
                },
                'ResourceRecords': [{
                    'Value': '2.2.3.4',
                }],
                'TTL': 61,
                'HealthCheckId': '42',
            }],
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_resource_record_sets',
                             list_resource_record_sets_resp,
                             {'HostedZoneId': 'z42'})
        stubber.add_response('list_health_checks', {
            'HealthChecks': [{
                'Id': '42',
                'CallerReference': self.caller_ref,
                'HealthCheckConfig': {
                    'Type': 'HTTPS',
                    'FullyQualifiedDomainName': 'a.unit.tests',
                    'IPAddress': '2.2.3.4',
                    'ResourcePath': '/_dns',
                    'Type': 'HTTPS',
                    'Port': 443,
                    'MeasureLatency': True
                },
                'HealthCheckVersion': 2,
            }],
            'IsTruncated': False,
            'MaxItems': '100',
            'Marker': '',
        })
        extra = provider._extra_changes(desired=desired, changes=[])
        self.assertEquals(0, len(extra))
        stubber.assert_no_pending_responses()

        # change b/c of healthcheck path
        record._octodns['healthcheck'] = {
            'path': '/_ready'
        }
        extra = provider._extra_changes(desired=desired, changes=[])
        self.assertEquals(1, len(extra))
        stubber.assert_no_pending_responses()

    def test_extra_change_dynamic_has_health_check(self):
        provider, stubber = self._get_stubbed_provider()

        list_hosted_zones_resp = {
            'HostedZones': [{
                'Name': 'unit.tests.',
                'Id': 'z42',
                'CallerReference': 'abc',
            }],
            'Marker': 'm',
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_hosted_zones', list_hosted_zones_resp, {})

        # record with geo and no health check returns change
        desired = Zone('unit.tests.', [])
        record = Record.new(desired, 'a', {
            'ttl': 30,
            'type': 'A',
            'value': '1.2.3.4',
            'dynamic': {
                'pools': {
                    'one': {
                        'values': [{
                            'value': '2.2.3.4',
                        }],
                    },
                },
                'rules': [{
                    'pool': 'one',
                }],
            },
        })
        desired.add_record(record)
        list_resource_record_sets_resp = {
            'ResourceRecordSets': [{
                # other name
                'Name': 'unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'CountryCode': '*',
                },
                'ResourceRecords': [{
                    'Value': '1.2.3.4',
                }],
                'TTL': 61,
            }, {
                # matching name, other type
                'Name': 'a.unit.tests.',
                'Type': 'AAAA',
                'ResourceRecords': [{
                    'Value': '2001:0db8:3c4d:0015:0000:0000:1a2f:1a4b'
                }],
                'TTL': 61,
            }, {
                # default value pool
                'Name': '_octodns-default-pool.a.unit.tests.',
                'Type': 'A',
                'GeoLocation': {
                    'CountryCode': '*',
                },
                'ResourceRecords': [{
                    'Value': '1.2.3.4',
                }],
                'TTL': 61,
            }, {
                # match
                'Name': '_octodns-one-value.a.unit.tests.',
                'Type': 'A',
                'ResourceRecords': [{
                    'Value': '2.2.3.4',
                }],
                'TTL': 61,
                'HealthCheckId': '42',
            }],
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_resource_record_sets',
                             list_resource_record_sets_resp,
                             {'HostedZoneId': 'z42'})
        stubber.add_response('list_health_checks', {
            'HealthChecks': [{
                'Id': '42',
                'CallerReference': self.caller_ref,
                'HealthCheckConfig': {
                    'Type': 'HTTPS',
                    'FullyQualifiedDomainName': 'a.unit.tests',
                    'IPAddress': '2.2.3.4',
                    'ResourcePath': '/_dns',
                    'Type': 'HTTPS',
                    'Port': 443,
                    'MeasureLatency': True
                },
                'HealthCheckVersion': 2,
            }],
            'IsTruncated': False,
            'MaxItems': '100',
            'Marker': '',
        })
        extra = provider._extra_changes(desired=desired, changes=[])
        self.assertEquals(0, len(extra))
        stubber.assert_no_pending_responses()

        # change b/c of healthcheck path
        record._octodns['healthcheck'] = {
            'path': '/_ready'
        }
        extra = provider._extra_changes(desired=desired, changes=[])
        self.assertEquals(1, len(extra))
        stubber.assert_no_pending_responses()

        # change b/c of healthcheck host
        record._octodns['healthcheck'] = {
            'host': 'foo.bar.io'
        }
        extra = provider._extra_changes(desired=desired, changes=[])
        self.assertEquals(1, len(extra))
        stubber.assert_no_pending_responses()

    def _get_test_plan(self, max_changes):

        provider = Route53Provider('test', 'abc', '123', max_changes)

        # Use the stubber
        stubber = Stubber(provider._conn)
        stubber.activate()

        got = Zone('unit.tests.', [])

        list_hosted_zones_resp = {
            'HostedZones': [],
            'Marker': 'm',
            'IsTruncated': False,
            'MaxItems': '100',
        }
        stubber.add_response('list_hosted_zones', list_hosted_zones_resp,
                             {})

        create_hosted_zone_resp = {
            'HostedZone': {
                'Name': 'unit.tests.',
                'Id': 'z42',
                'CallerReference': 'abc',
            },
            'ChangeInfo': {
                'Id': 'a12',
                'Status': 'PENDING',
                'SubmittedAt': '2017-01-29T01:02:03Z',
                'Comment': 'hrm',
            },
            'DelegationSet': {
                'Id': 'b23',
                'CallerReference': 'blip',
                'NameServers': [
                    'n12.unit.tests.',
                ],
            },
            'Location': 'us-east-1',
        }
        stubber.add_response('create_hosted_zone',
                             create_hosted_zone_resp, {
                                 'Name': got.name,
                                 'CallerReference': ANY,
                             })

        stubber.add_response('list_health_checks',
                             {
                                 'HealthChecks': self.health_checks,
                                 'IsTruncated': False,
                                 'MaxItems': '100',
                                 'Marker': '',
                             })

        stubber.add_response('change_resource_record_sets',
                             {'ChangeInfo': {
                                 'Id': 'id',
                                 'Status': 'PENDING',
                                 'SubmittedAt': '2017-01-29T01:02:03Z',
                             }}, {'HostedZoneId': 'z42', 'ChangeBatch': ANY})

        plan = provider.plan(self.expected)

        return provider, plan

    # _get_test_plan() returns a plan with 11 modifications, 17 RRs

    @patch('octodns.provider.route53.Route53Provider._load_records')
    @patch('octodns.provider.route53.Route53Provider._really_apply')
    def test_apply_1(self, really_apply_mock, _):

        # 18 RRs with max of 19 should only get applied in one call
        provider, plan = self._get_test_plan(19)
        provider.apply(plan)
        really_apply_mock.assert_called_once()

    @patch('octodns.provider.route53.Route53Provider._load_records')
    @patch('octodns.provider.route53.Route53Provider._really_apply')
    def test_apply_2(self, really_apply_mock, _):

        # 18 RRs with max of 17 should only get applied in two calls
        provider, plan = self._get_test_plan(18)
        provider.apply(plan)
        self.assertEquals(2, really_apply_mock.call_count)

    @patch('octodns.provider.route53.Route53Provider._load_records')
    @patch('octodns.provider.route53.Route53Provider._really_apply')
    def test_apply_3(self, really_apply_mock, _):

        # with a max of seven modifications, four calls
        provider, plan = self._get_test_plan(7)
        provider.apply(plan)
        self.assertEquals(4, really_apply_mock.call_count)

    @patch('octodns.provider.route53.Route53Provider._load_records')
    @patch('octodns.provider.route53.Route53Provider._really_apply')
    def test_apply_4(self, really_apply_mock, _):

        # with a max of 11 modifications, two calls
        provider, plan = self._get_test_plan(11)
        provider.apply(plan)
        self.assertEquals(2, really_apply_mock.call_count)

    @patch('octodns.provider.route53.Route53Provider._load_records')
    @patch('octodns.provider.route53.Route53Provider._really_apply')
    def test_apply_bad(self, really_apply_mock, _):

        # with a max of 1 modifications, fail
        provider, plan = self._get_test_plan(1)
        with self.assertRaises(Exception) as ctx:
            provider.apply(plan)
        self.assertTrue('modifications' in ctx.exception.message)

    def test_semicolon_fixup(self):
        provider = Route53Provider('test', 'abc', '123')

        self.assertEquals({
            'type': 'TXT',
            'ttl': 30,
            'values': [
                'abcd\\; ef\\;g',
                'hij\\; klm\\;n',
            ],
        }, provider._data_for_quoted({
            'ResourceRecords': [{
                'Value': '"abcd; ef;g"',
            }, {
                'Value': '"hij\\; klm\\;n"',
            }],
            'TTL': 30,
            'Type': 'TXT',
        }))

    def test_client_max_attempts(self):
        provider = Route53Provider('test', 'abc', '123',
                                   client_max_attempts=42)
        # NOTE: this will break if boto ever changes the impl details...
        self.assertEquals(43, provider._conn.meta.events
                          ._unique_id_handlers['retry-config-route53']
                          ['handler']._checker.__dict__['_max_attempts'])

    def test_data_for_dynamic(self):
        provider = Route53Provider('test', 'abc', '123')

        data = provider._data_for_dynamic('', 'A', dynamic_rrsets)
        self.assertEquals(dynamic_record_data, data)

    @patch('octodns.provider.route53.Route53Provider._get_zone_id')
    @patch('octodns.provider.route53.Route53Provider._load_records')
    def test_dynamic_populate(self, load_records_mock, get_zone_id_mock):
        provider = Route53Provider('test', 'abc', '123')

        get_zone_id_mock.side_effect = ['z44']
        load_records_mock.side_effect = [dynamic_rrsets]

        got = Zone('unit.tests.', [])
        provider.populate(got)

        self.assertEquals(1, len(got.records))
        record = list(got.records)[0]
        self.assertEquals('', record.name)
        self.assertEquals('A', record._type)
        self.assertEquals([
            '1.1.2.1',
            '1.1.2.2',
        ], record.values)
        self.assertTrue(record.dynamic)

        self.assertEquals({
            'ap-southeast-1': {
                'fallback': 'us-east-1',
                'values': [{
                    'weight': 2, 'value': '1.4.1.1'
                }, {
                    'weight': 2, 'value': '1.4.1.2'
                }]
            },
            'eu-central-1': {
                'fallback': 'us-east-1',
                'values': [{
                    'weight': 1, 'value': '1.3.1.1'
                }, {
                    'weight': 1, 'value': '1.3.1.2'
                }],
            },
            'us-east-1': {
                'fallback': None,
                'values': [{
                    'weight': 1, 'value': '1.5.1.1'
                }, {
                    'weight': 1, 'value': '1.5.1.2'
                }],
            }
        }, {k: v.data for k, v in record.dynamic.pools.items()})

        self.assertEquals([
            {
                'geos': ['AS-CN', 'AS-JP'],
                'pool': 'ap-southeast-1',
            }, {
                'geos': ['EU', 'NA-US-FL'],
                'pool': 'eu-central-1',
            }, {
                'pool': 'us-east-1',
            }], [r.data for r in record.dynamic.rules])


class DummyProvider(object):

    def get_health_check_id(self, *args, **kwargs):
        return None


class TestRoute53Records(TestCase):
    existing = Zone('unit.tests.', [])
    record_a = Record.new(existing, '', {
        'geo': {
            'NA-US': ['2.2.2.2', '3.3.3.3'],
            'OC': ['4.4.4.4', '5.5.5.5']
        },
        'ttl': 99,
        'type': 'A',
        'values': ['9.9.9.9']
    })

    def test_value_fors(self):
        route53_record = _Route53Record(None, self.record_a, False)

        for value in (None, '', 'foo', 'bar', '1.2.3.4'):
            converted = route53_record._value_convert_value(value,
                                                            self.record_a)
            self.assertEquals(value, converted)

        record_txt = Record.new(self.existing, 'txt', {
            'ttl': 98,
            'type': 'TXT',
            'value': 'Not Important',
        })

        # We don't really have to test the details fo chunked_value as that's
        # tested elsewhere, we just need to make sure that it's plumbed up and
        # working
        self.assertEquals('"Not Important"', route53_record
                          ._value_convert_quoted(record_txt.values[0],
                                                 record_txt))

    def test_route53_record(self):
        a = _Route53Record(None, self.record_a, False)
        self.assertEquals(a, a)
        b = _Route53Record(None, Record.new(self.existing, '',
                                            {'ttl': 32, 'type': 'A',
                                             'values': ['8.8.8.8',
                                                        '1.1.1.1']}),
                           False)
        self.assertEquals(b, b)
        c = _Route53Record(None, Record.new(self.existing, 'other',
                                            {'ttl': 99, 'type': 'A',
                                             'values': ['9.9.9.9']}),
                           False)
        self.assertEquals(c, c)
        d = _Route53Record(None, Record.new(self.existing, '',
                                            {'ttl': 42, 'type': 'MX',
                                             'value': {
                                                 'preference': 10,
                                                 'exchange': 'foo.bar.'}}),
                           False)
        self.assertEquals(d, d)

        # Same fqdn & type is same record
        self.assertEquals(a, b)
        # Same name & different type is not the same
        self.assertNotEquals(a, d)
        # Different name & same type is not the same
        self.assertNotEquals(a, c)

        # Same everything, different class is not the same
        e = _Route53GeoDefault(None, self.record_a, False)
        self.assertNotEquals(a, e)

        provider = DummyProvider()
        f = _Route53GeoRecord(provider, self.record_a, 'NA-US',
                              self.record_a.geo['NA-US'], False)
        self.assertEquals(f, f)
        g = _Route53GeoRecord(provider, self.record_a, 'OC',
                              self.record_a.geo['OC'], False)
        self.assertEquals(g, g)

        # Geo and non-geo are not the same, using Geo as primary to get it's
        # __cmp__
        self.assertNotEquals(f, a)
        # Same everything, different geo's is not the same
        self.assertNotEquals(f, g)

        # Make sure it doesn't blow up
        a.__repr__()
        e.__repr__()
        f.__repr__()

    def test_dynamic_value_delete(self):
        provider = DummyProvider()
        geo = _Route53DynamicValue(provider, self.record_a, 'iad', '2.2.2.2',
                                   1, 0, False)

        rrset = {
            'HealthCheckId': 'x12346z',
            'Name': '_octodns-iad-value.unit.tests.',
            'ResourceRecords': [{
                'Value': '2.2.2.2'
            }],
            'SetIdentifier': 'iad-000',
            'TTL': 99,
            'Type': 'A',
            'Weight': 1,
        }

        candidates = [
            # Empty, will test no SetIdentifier
            {},
            # Non-matching
            {
                'SetIdentifier': 'not-a-match',
            },
            # Same set-id, different name
            {
                'Name': 'not-a-match',
                'SetIdentifier': 'x12346z',
            },
            rrset,
        ]

        # Provide a matching rrset so that we'll just use it for the delete
        # rathr than building up an almost identical one, note the way we'll
        # know that we got the one we passed in is that it'll have a
        # HealthCheckId and one that was created wouldn't since DummyProvider
        # stubs out the lookup for them
        mod = geo.mod('DELETE', candidates)
        self.assertEquals('x12346z', mod['ResourceRecordSet']['HealthCheckId'])

        # If we don't provide the candidate rrsets we get back exactly what we
        # put in minus the healthcheck
        rrset['HealthCheckId'] = None
        mod = geo.mod('DELETE', [])
        self.assertEquals(rrset, mod['ResourceRecordSet'])

    def test_geo_delete(self):
        provider = DummyProvider()
        geo = _Route53GeoRecord(provider, self.record_a, 'NA-US',
                                self.record_a.geo['NA-US'], False)

        rrset = {
            'GeoLocation': {
                'CountryCode': 'US'
            },
            'HealthCheckId': 'x12346z',
            'Name': 'unit.tests.',
            'ResourceRecords': [{
                'Value': '2.2.2.2'
            }, {
                'Value': '3.3.3.3'
            }],
            'SetIdentifier': 'NA-US',
            'TTL': 99,
            'Type': 'A'
        }

        candidates = [
            # Empty, will test no SetIdentifier
            {},
            {
                'SetIdentifier': 'not-a-match',
            },
            # Same set-id, different name
            {
                'Name': 'not-a-match',
                'SetIdentifier': 'x12346z',
            },
            rrset,
        ]

        # Provide a matching rrset so that we'll just use it for the delete
        # rathr than building up an almost identical one, note the way we'll
        # know that we got the one we passed in is that it'll have a
        # HealthCheckId and one that was created wouldn't since DummyProvider
        # stubs out the lookup for them
        mod = geo.mod('DELETE', candidates)
        self.assertEquals('x12346z', mod['ResourceRecordSet']['HealthCheckId'])

        # If we don't provide the candidate rrsets we get back exactly what we
        # put in minus the healthcheck
        del rrset['HealthCheckId']
        mod = geo.mod('DELETE', [])
        self.assertEquals(rrset, mod['ResourceRecordSet'])

    def test_new_dynamic(self):
        provider = Route53Provider('test', 'abc', '123')

        # Just so boto won't try and make any calls
        stubber = Stubber(provider._conn)
        stubber.activate()

        # We'll assume we create all healthchecks here, this functionality is
        # thoroughly tested elsewhere
        provider._health_checks = {}
        # When asked for a healthcheck return dummy info
        provider.get_health_check_id = lambda r, v, c: 'hc42'

        zone = Zone('unit.tests.', [])
        record = Record.new(zone, '', dynamic_record_data)

        # Convert a record into _Route53Records
        route53_records = _Route53Record.new(provider, record, 'z45',
                                             creating=True)
        self.assertEquals(18, len(route53_records))

        # Convert the route53_records into mods
        self.assertEquals([{
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'HealthCheckId': 'hc42',
                'Name': '_octodns-ap-southeast-1-value.unit.tests.',
                'ResourceRecords': [{
                    'Value': '1.4.1.2'}],
                'SetIdentifier': 'ap-southeast-1-001',
                'TTL': 60,
                'Type': 'A',
                'Weight': 2
            }
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'HealthCheckId': 'hc42',
                'Name': '_octodns-ap-southeast-1-value.unit.tests.',
                'ResourceRecords': [{
                    'Value': '1.4.1.1'}],
                'SetIdentifier': 'ap-southeast-1-000',
                'TTL': 60,
                'Type': 'A',
                'Weight': 2
            }
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-ap-southeast-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'
                },
                'GeoLocation': {
                    'CountryCode': 'JP'},
                'Name': 'unit.tests.',
                'SetIdentifier': '0-ap-southeast-1-AS-JP',
                'Type': 'A'
            }
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-eu-central-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'},
                'GeoLocation': {
                    'CountryCode': 'US',
                    'SubdivisionCode': 'FL',
                },
                'Name': 'unit.tests.',
                'SetIdentifier': '1-eu-central-1-NA-US-FL',
                'Type': 'A'}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-us-east-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'},
                'GeoLocation': {
                    'CountryCode': '*'},
                'Name': 'unit.tests.',
                'SetIdentifier': '2-us-east-1-None',
                'Type': 'A'}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-us-east-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'},
                'Failover': 'SECONDARY',
                'Name': '_octodns-ap-southeast-1-pool.unit.tests.',
                'SetIdentifier': 'ap-southeast-1-Secondary-us-east-1',
                'Type': 'A'}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-ap-southeast-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'},
                'GeoLocation': {
                    'CountryCode': 'CN'},
                'Name': 'unit.tests.',
                'SetIdentifier': '0-ap-southeast-1-AS-CN',
                'Type': 'A'}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-us-east-1-value.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'},
                'Failover': 'PRIMARY',
                'Name': '_octodns-us-east-1-pool.unit.tests.',
                'SetIdentifier': 'us-east-1-Primary',
                'Type': 'A'}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-eu-central-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'},
                'GeoLocation': {
                    'ContinentCode': 'EU'},
                'Name': 'unit.tests.',
                'SetIdentifier': '1-eu-central-1-EU',
                'Type': 'A'}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-eu-central-1-value.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'},
                'Failover': 'PRIMARY',
                'Name': '_octodns-eu-central-1-pool.unit.tests.',
                'SetIdentifier': 'eu-central-1-Primary',
                'Type': 'A'}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'Name': '_octodns-default-pool.unit.tests.',
                'ResourceRecords': [{
                    'Value': '1.1.2.1'},
                    {
                        'Value': '1.1.2.2'}],
                'TTL': 60,
                'Type': 'A'}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'HealthCheckId': 'hc42',
                'Name': '_octodns-eu-central-1-value.unit.tests.',
                'ResourceRecords': [{
                    'Value': '1.3.1.2'}],
                'SetIdentifier': 'eu-central-1-001',
                'TTL': 60,
                'Type': 'A',
                'Weight': 1}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'HealthCheckId': 'hc42',
                'Name': '_octodns-eu-central-1-value.unit.tests.',
                'ResourceRecords': [{
                    'Value': '1.3.1.1'}],
                'SetIdentifier': 'eu-central-1-000',
                'TTL': 60,
                'Type': 'A',
                'Weight': 1}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-default-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'},
                'Failover': 'SECONDARY',
                'Name': '_octodns-us-east-1-pool.unit.tests.',
                'SetIdentifier': 'us-east-1-Secondary-default',
                'Type': 'A'}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'HealthCheckId': 'hc42',
                'Name': '_octodns-us-east-1-value.unit.tests.',
                'ResourceRecords': [{
                    'Value': '1.5.1.2'}],
                'SetIdentifier': 'us-east-1-001',
                'TTL': 60,
                'Type': 'A',
                'Weight': 1}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'HealthCheckId': 'hc42',
                'Name': '_octodns-us-east-1-value.unit.tests.',
                'ResourceRecords': [{
                    'Value': '1.5.1.1'}],
                'SetIdentifier': 'us-east-1-000',
                'TTL': 60,
                'Type': 'A',
                'Weight': 1}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-ap-southeast-1-value.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'},
                'Failover': 'PRIMARY',
                'Name': '_octodns-ap-southeast-1-pool.unit.tests.',
                'SetIdentifier': 'ap-southeast-1-Primary',
                'Type': 'A'}
        }, {
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'AliasTarget': {
                    'DNSName': '_octodns-us-east-1-pool.unit.tests.',
                    'EvaluateTargetHealth': True,
                    'HostedZoneId': 'z45'},
                'Failover': 'SECONDARY',
                'Name': '_octodns-eu-central-1-pool.unit.tests.',
                'SetIdentifier': 'eu-central-1-Secondary-us-east-1',
                'Type': 'A'}
        }], [r.mod('CREATE', []) for r in route53_records])

        for route53_record in route53_records:
            # Smoke test stringification
            route53_record.__repr__()


class TestModKeyer(TestCase):

    def test_mod_keyer(self):

        # First "column"

        # Deletes come first
        self.assertEquals((0, 0, 'something'), _mod_keyer({
            'Action': 'DELETE',
            'ResourceRecordSet': {
                'Name': 'something',
            }
        }))

        # Creates come next
        self.assertEquals((1, 0, 'another'), _mod_keyer({
            'Action': 'CREATE',
            'ResourceRecordSet': {
                'Name': 'another',
            }
        }))

        # Then upserts
        self.assertEquals((2, 0, 'last'), _mod_keyer({
            'Action': 'UPSERT',
            'ResourceRecordSet': {
                'Name': 'last',
            }
        }))

        # Second "column" value records tested above

        # AliasTarget primary second (to value)
        self.assertEquals((0, 1, 'thing'), _mod_keyer({
            'Action': 'DELETE',
            'ResourceRecordSet': {
                'AliasTarget': 'some-target',
                'Failover': 'PRIMARY',
                'Name': 'thing',
            }
        }))

        # AliasTarget secondary third
        self.assertEquals((0, 2, 'thing'), _mod_keyer({
            'Action': 'DELETE',
            'ResourceRecordSet': {
                'AliasTarget': 'some-target',
                'Failover': 'SECONDARY',
                'Name': 'thing',
            }
        }))

        # GeoLocation fourth
        self.assertEquals((0, 3, 'some-id'), _mod_keyer({
            'Action': 'DELETE',
            'ResourceRecordSet': {
                'GeoLocation': 'some-target',
                'SetIdentifier': 'some-id',
            }
        }))

        # The third "column" has already been tested above, Name/SetIdentifier
